from __future__ import annotations

import logging
from abc import abstractmethod, abstractstaticmethod
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Hashable, TYPE_CHECKING, overload
import warnings
from psygnal import SignalGroup, Signal

from tabulous.widgets import _doc, _component as _comp
from tabulous.widgets._keymap_abc import SupportKeyMap
from tabulous.widgets._source import Source
from tabulous.types import ItemInfo, HeaderInfo, EvalInfo
from tabulous._psygnal import SignalArray, InCellRangedSlot

if TYPE_CHECKING:
    from typing_extensions import Self, Literal
    import numpy as np
    import pandas as pd
    from collections_undo import UndoManager
    from qtpy import QtWidgets as QtW
    from magicgui.widgets import Widget

    from tabulous._qt import QTableLayer, QSpreadSheet, QTableGroupBy, QTableDisplay
    from tabulous._qt._table import QBaseTable
    from tabulous._qt._table._base._overlay import QOverlayFrame

    LayoutString = Literal["horizontal", "vertical"]

logger = logging.getLogger("tabulous")


class TableSignals(SignalGroup):
    """Signal group for a Table."""

    data = SignalArray(ItemInfo)
    index = Signal(HeaderInfo)
    columns = Signal(HeaderInfo)
    evaluated = Signal(EvalInfo)
    selections = Signal(_comp.SelectionRanges)
    renamed = Signal(str)


class ViewMode(Enum):
    """Enum of view modes"""

    normal = "normal"
    horizontal = "horizontal"
    vertical = "vertical"
    popup = "popup"

    def __eq__(self, other):
        if isinstance(other, str):
            return self.name == other
        return super().__eq__(other)

    def __repr__(self) -> str:
        return f"<{type(self).__name__}.{self.name}>"


# #####################################################################
#   The abstract base class of tables.
# #####################################################################


class TableBase(SupportKeyMap):
    """The base class for a table layer."""

    _Default_Name = "None"

    cell = _comp.CellInterface()
    index = _comp.VerticalHeaderInterface()
    columns = _comp.HorizontalHeaderInterface()
    plt = _comp.PlotInterface()
    proxy = _comp.ProxyInterface()
    text_color = _comp.TextColormapInterface()
    background_color = _comp.BackgroundColormapInterface()
    formatter = _comp.TextFormatterInterface()
    validator = _comp.ValidatorInterface()
    selections = _comp.SelectionRanges()
    highlights = _comp.HighlightRanges()
    loc = _comp.TableLocIndexer()
    iloc = _comp.TableILocIndexer()

    def __init__(
        self,
        data: Any = None,
        name: str | None = None,
        editable: bool = True,
        metadata: dict[str, Any] | None = None,
    ):
        from tabulous._qt import get_app
        from tabulous._map_model import SlotRefMapping

        _ = get_app()

        _data = self._normalize_data(data)

        if name is None:
            name = self._Default_Name
        self.events = TableSignals()
        self._name = str(name)
        self._qwidget = self._create_backend(_data)
        self._install_actions()
        self._qwidget.connectSelectionChangedSignal(self._emit_selections)

        self._qwidget._qtable_view._table_map = SlotRefMapping(self)
        self._view_mode = ViewMode.normal
        self._metadata: dict[str, Any] = metadata or {}
        self._source = Source()

        if self.mutable:
            with self._qwidget._mgr.blocked():
                self._qwidget.setEditable(editable)
            self.native.connectItemChangedSignal(
                self._emit_data_changed_signal,
                self.events.index.emit,
                self.events.columns.emit,
                self.events.evaluated.emit,
            )

            self.events.evaluated.connect(self._emit_evaluated)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}<{self.name!r}>"

    # fmt: off
    @overload
    def __getitem__(self, key: str) -> _comp.TableSeries: ...
    @overload
    def __getitem__(self, key: list[str]) -> _comp.TableSubset: ...
    # fmt: on

    def __getitem__(self, key):
        if isinstance(key, str):
            return _comp.TableSeries(self, slice(None), key)
        elif isinstance(key, list):
            return _comp.TableSubset(self, slice(None), key)
        raise TypeError(f"Invalid key type: {type(key)}")

    @property
    def cellref(self):
        warnings.warn(
            "table.cellref is deprecated. Use table.cell.ref instead.",
            DeprecationWarning,
        )
        return self.cell.ref

    @property
    def source(self) -> Source:
        """The source of the table."""
        return self._source

    def _emit_data_changed_signal(self, info: ItemInfo) -> None:
        r, c = info.row, info.column
        if info.value is info.DELETED or info.old_value is info.INSERTED:
            # insertion/deletion emits signal from the next row/column.
            _is_deleted = info.value is info.DELETED
            if r == slice(None):
                # column is deleted/inserted
                if isinstance(c, slice):
                    index, count = c.start, c.stop - c.start
                else:
                    index, count = c, 1
                if _is_deleted:
                    self.events.data.remove_columns(index, count)
                else:
                    self.events.data.insert_columns(index, count)
                self.events.data[:, c.start :].emit(info)
            else:
                # row is deleted/inserted
                if isinstance(r, slice):
                    index, count = r.start, r.stop - r.start
                else:
                    index, count = r, 1
                if _is_deleted:
                    self.events.data.remove_rows(index, count)
                else:
                    self.events.data.insert_rows(index, count)
                self.events.data[r.start :, :].emit(info)
        else:
            self.events.data[r, c].emit(info)

    @abstractmethod
    def _create_backend(self, data: pd.DataFrame) -> QBaseTable:
        """This function creates a backend widget."""

    @abstractstaticmethod
    def _normalize_data(self, data):
        """Data normalization before setting a new data."""

    @property
    def table_type(self) -> str:
        """Return the table type in string."""
        return type(self).__name__

    @property
    def data(self) -> pd.DataFrame:
        """Table data."""
        return self._qwidget.getDataFrame()

    @data.setter
    def data(self, value: Any):
        """Set table data."""
        _data = self._normalize_data(value)
        self._qwidget.setDataFrame(_data)

    @property
    def data_shown(self) -> pd.DataFrame:
        """Return the data shown in the table (filter considered)."""
        return self._qwidget.dataShown(parse=True)

    @property
    def mutable(self):
        """Mutability of the table type."""
        from tabulous._qt._table import QMutableTable

        return isinstance(self._qwidget, QMutableTable)

    @property
    def table_shape(self) -> tuple[int, int]:
        """Shape of table."""
        return self._qwidget.tableShape()

    @property
    def metadata(self) -> dict[str, Any]:
        """Metadata of the table."""
        return self._metadata

    @metadata.setter
    def metadata(self, value: dict[str, Any]) -> None:
        """Set metadata of the table."""
        if not isinstance(value, dict):
            raise TypeError("metadata must be a dict")
        self._metadata = value

    @property
    def zoom(self) -> float:
        """Zoom factor of table."""
        return self._qwidget.zoom()

    @zoom.setter
    def zoom(self, value: float):
        """Set zoom factor of table."""
        return self._qwidget.setZoom(value)

    @property
    def name(self) -> str:
        """Table name."""
        return self._name

    @name.setter
    def name(self, value: str):
        """Set table name."""
        self._name = str(value)
        return self.events.renamed.emit(self._name)

    @property
    def editable(self) -> bool:
        """Editability of table."""
        if self.mutable:
            return self._qwidget.isEditable()
        else:
            return False

    @editable.setter
    def editable(self, value: bool):
        """Set editability of table."""
        if self.mutable:
            self._qwidget.setEditable(value)
        elif value:
            raise ValueError("Table is not mutable.")

    @property
    def native(self) -> QBaseTable:
        """The backend widget."""
        return self._qwidget

    @property
    def filter(self):
        """Filter applied to the table."""
        warnings.warn(
            "'filter' is deprecated. Please use `table.proxy.filter` API instead",
            DeprecationWarning,
        )
        prx = self.proxy._get_proxy_object()
        if prx.proxy_type == "filter":
            return prx._obj
        return None

    @filter.setter
    def filter(self, val):
        warnings.warn(
            "'filter' is deprecated. Please use `table.proxy.filter` API instead",
            DeprecationWarning,
        )
        return self.proxy.filter(val)

    def refresh(self) -> None:
        """Refresh the table view and force table to update."""
        return self._qwidget.refreshTable(process=True)

    def move_loc(self, row: Hashable, column: Hashable):
        """
        Move to a location in the table using axis label.
        >>> table.move_loc("index-2", "column-4")
        """
        data = self.data
        r = data.index.get_loc(row)
        c = data.columns.get_loc(column)
        return self._qwidget.moveToItem(r, c)

    def move_iloc(self, row: int | None, column: int | None, scroll: bool = True):
        """
        Move to a location in the table using indices.
        >>> table.move_iloc(2, 4)
        """
        shape = self.table_shape
        row_outofrange = False if row is None else row >= shape[0]
        col_outofrange = False if column is None else column >= shape[1]
        if row_outofrange or col_outofrange:
            raise IndexError(
                f"Indices {(row, column)!r} out of range of table shape {shape!r}."
            )
        self._qwidget.moveToItem(row, column)
        if scroll:
            qtable_view = self._qwidget._qtable_view
            index = qtable_view.model().index(row, column)
            qtable_view.scrollTo(index)
        return None

    text_formatter = formatter  # alias

    @property
    def view_mode(self) -> ViewMode:
        """View mode of the table."""
        return self._view_mode

    @view_mode.setter
    def view_mode(self, mode) -> None:
        """Set view mode of the table."""

        if self._qwidget.parentViewer() is None:
            raise ValueError(
                "Table is not attached to a viewer. Cannot change the view mode."
            )
        if mode is None:
            mode = ViewMode.normal
        else:
            mode = ViewMode(mode)

        if mode in (ViewMode.horizontal, ViewMode.vertical):
            self._qwidget.setDualView(orientation=mode.name)
        elif mode == ViewMode.popup:
            view = self._qwidget.setPopupView()
            view.popup.setTitle(self.name)

            @view.popup.closed.connect
            def _():
                self.view_mode = ViewMode.normal
                self._qwidget._qtable_view.setFocus()

        elif mode == ViewMode.normal:
            self._qwidget.resetViewMode()
        else:
            raise ValueError(f"Unknown view mode: {mode!r}.")

        self._view_mode = mode
        return None

    @property
    def layout(self) -> LayoutString:
        """Table layout."""
        if self._qwidget.orientation() == 0:
            return "vertical"
        elif self._qwidget.orientation() == 1:
            return "horizontal"
        else:
            raise ValueError("Unknown layout.")

    @layout.setter
    def layout(self, layout: LayoutString) -> None:
        """Set table layout."""
        if layout == "vertical":
            self._qwidget.setOrientation(0)
        elif layout == "horizontal":
            self._qwidget.setOrientation(1)
        else:
            raise ValueError(f"Unknown layout: {layout!r}.")

    @property
    def undo_manager(self) -> UndoManager:
        """Return the undo manager."""
        return self._qwidget._mgr

    def add_side_widget(self, widget: QtW.QWidget | Widget, *, name: str = ""):
        """
        Add a side widget to the table.

        A side widget is a widget that is docked to the side area of the table.
        It is visible only when the table is active. Thus, it is a good place
        to add a widget specific to the table.

        Parameters
        ----------
        widget: QWidget or magicgui Widget
            The widget to add.
        name : str, optional
            Name of the size widget. Use the input widget name by default.
        """
        if hasattr(widget, "native"):
            name = name or widget.name
            widget = widget.native
        else:
            name = name or widget.objectName()

        self._qwidget.addSideWidget(widget, name=name)
        return widget

    def add_overlay_widget(
        self,
        widget: QtW.QWidget | Widget,
        *,
        label: str = "",
        topleft: tuple[float, float] = (0, 0),
        size: tuple[float, float] | None = None,
        grip: bool = True,
    ) -> QOverlayFrame:
        """
        Add a widget overlaid over the table.

        An overlay widget is shown on top of the table, just like the chart in Excel.

        Parameters
        ----------
        widget: QWidget or magicgui Widget
            The widget to add.
        label : str, optional
            Label that is shown in the bottom of the widget.
        topleft: tuple of float, optional
            Top-left position of the widget described as the row and column indices.
        """
        if hasattr(widget, "native"):
            widget = widget.native

        return self._qwidget.addOverlayWidget(
            widget,
            label=label,
            topleft=topleft,
            size=size,
            grip=grip,
        )

    def save(self, path: str | Path) -> None:
        """Save table data to the given path."""
        from tabulous._io import save_file

        path = Path(path)
        save_file(path, self.data)
        self._source = self.source.replace(path=path)
        return None

    def screenshot(self) -> np.ndarray:
        """Get screenshot of the widget."""
        return self._qwidget.screenshot()

    def save_screenshot(self, path: str):
        """Save screenshot of the widget."""
        from PIL import Image

        arr = self.screenshot()
        Image.fromarray(arr).save(path)
        return None

    def _emit_selections(self):
        with self.selections.blocked():
            # Block selection to avoid recursive update.
            self.events.selections.emit(self.selections)
        return None

    def _emit_evaluated(self, info: EvalInfo):
        if info.expr == "":
            return None

        qtable = self.native
        qtable_view = qtable._qtable_view

        slot = InCellRangedSlot.from_table(self, info.expr, info.pos)

        def _raise(e: Exception):
            # the common exception handling
            if not isinstance(e, (SyntaxError, AttributeError)):
                # Update cell text with the exception object.
                try:
                    del qtable_view._focused_widget
                except RuntimeError:
                    pass
                with qtable_view._selection_model.blocked():
                    qtable.setDataFrameValue(*info.pos, "#ERROR")
                return None
            # SyntaxError/AttributeError might be caused by mistouching. Don't close
            # the editor.
            e.args = (f"{str(e)}\n>>> {info.expr}",)
            raise e

        if not info.is_ref:
            result = slot.evaluate()
            if e := result.get_err():
                _raise(e)
            else:
                self.move_iloc(*info.pos)
        else:
            if next(iter(slot.range), None) is None:
                # if no reference exists, evaluate the expression as "=..." form.
                return self._emit_evaluated(
                    EvalInfo(info.pos, info.source_pos, info.expr, False)
                )
            with qtable._mgr.merging(formatter=lambda cmds: cmds[-1].format()):
                # call here to properly update undo stack
                result = slot.evaluate()
                if e := result.get_err():
                    _raise(e)
                qtable.setInCellSlot(info.source_pos, slot)

        del qtable_view._focused_widget
        return None

    def _wrap_command(self, cmd: Callable):
        def _f(*_):
            logger.debug(f"Command: {cmd.__module__.split('.')[-1]}.{cmd.__name__}")
            if _qviewer := self._qwidget.parentViewer():
                _viewer = _qviewer._table_viewer
            else:
                from tabulous.widgets._mainwindow import DummyViewer

                _viewer = DummyViewer(self)
            return cmd(_viewer)

        return _f

    def _install_actions(self):
        from tabulous import commands as cmds

        _wrap = self._wrap_command

        # fmt: off
        _hheader_register = self.columns.register
        _hheader_register("Color > Set text colormap", _wrap(cmds.column.set_text_colormap))  # noqa: E501
        _hheader_register("Color > Reset text colormap", _wrap(cmds.column.reset_text_colormap))  # noqa: E501
        _hheader_register("Color > Set opacity to text colormap", _wrap(cmds.column.set_text_colormap_opacity))  # noqa: E501
        self._qwidget._qtable_view.horizontalHeader().addSeparator("Color")
        _hheader_register("Color > Set background colormap", _wrap(cmds.column.set_background_colormap))  # noqa: E501
        _hheader_register("Color > Reset background colormap", _wrap(cmds.column.reset_background_colormap))  # noqa: E501
        _hheader_register("Color > Set opacity to background colormap", _wrap(cmds.column.set_background_colormap_opacity))  # noqa: E501

        _hheader_register("Formatter > Set text formatter", _wrap(cmds.column.set_text_formatter))  # noqa: E501
        _hheader_register("Formatter > Reset text formatter", _wrap(cmds.column.reset_text_formatter))  # noqa: E501
        self._qwidget._qtable_view.horizontalHeader().addSeparator()
        _hheader_register("Sort", _wrap(cmds.selection.sort_by_columns))
        _hheader_register("Filter", _wrap(cmds.selection.filter_by_columns))
        self._qwidget._qtable_view.horizontalHeader().addSeparator()

        _cell_register = self.cell.register
        _cell_register("Copy")(_wrap(cmds.selection.copy_data_tab_separated))
        _cell_register("Copy as ... > Tab separated text", _wrap(cmds.selection.copy_data_tab_separated))  # noqa: E501
        _cell_register("Copy as ... > Tab separated text with headers", _wrap(cmds.selection.copy_data_with_header_tab_separated))  # noqa: E501
        _cell_register("Copy as ... > Comma separated text", _wrap(cmds.selection.copy_data_comma_separated))  # noqa: E501
        _cell_register("Copy as ... > Comma separated text with headers", _wrap(cmds.selection.copy_data_with_header_comma_separated))  # noqa: E501
        self._qwidget.addSeparator("Copy as ...")
        _cell_register("Copy as ... > Markdown", _wrap(cmds.selection.copy_as_markdown))  # noqa: E501
        _cell_register("Copy as ... > reStructuredText grid table", _wrap(cmds.selection.copy_as_rst_grid))  # noqa: E501
        _cell_register("Copy as ... > reStructuredText simple table", _wrap(cmds.selection.copy_as_rst_simple))  # noqa: E501
        _cell_register("Copy as ... > Latex", _wrap(cmds.selection.copy_as_latex))
        _cell_register("Copy as ... > HTML", _wrap(cmds.selection.copy_as_html))
        _cell_register("Copy as ... > Literal", _wrap(cmds.selection.copy_as_literal))
        self._qwidget.addSeparator("Copy as ... ")
        _cell_register("Copy as ... > New table", _wrap(cmds.selection.copy_as_new_table))  # noqa: E501
        _cell_register("Copy as ... > New spreadsheet", _wrap(cmds.selection.copy_as_new_spreadsheet))  # noqa: E501
        _cell_register("Paste", _wrap(cmds.selection.paste_data_tab_separated))
        _cell_register("Paste from ... > Tab separated text", _wrap(cmds.selection.paste_data_tab_separated))  # noqa: E501
        _cell_register("Paste from ... > Comma separated text", _wrap(cmds.selection.paste_data_comma_separated))  # noqa: E501
        _cell_register("Paste from ... > numpy-style text", _wrap(cmds.selection.paste_data_from_numpy_string))  # noqa: E501
        _cell_register("Paste from ... > Markdown text", _wrap(cmds.selection.paste_data_from_markdown))  # noqa: E501
        self._qwidget.addSeparator()
        _cell_register("Sort in-place", _wrap(cmds.selection.sort_inplace))
        self._qwidget.addSeparator()
        _cell_register("Code ... > Data-changed signal", _wrap(cmds.selection.write_data_signal_in_console))  # noqa: E501
        _cell_register("Code ... > Get slice", _wrap(cmds.selection.write_slice_in_console))  # noqa: E501
        _cell_register("Add highlight", _wrap(cmds.selection.add_highlight))
        _cell_register("Delete highlight", _wrap(cmds.selection.delete_selected_highlight))  # noqa: E501
        self._qwidget.addSeparator()
        # fmt: on


# #############################################################################
#   Concrete table widgets
# #############################################################################


class _DataFrameTableLayer(TableBase):
    """Table layer for DataFrame."""

    _qwidget: QTableLayer | QSpreadSheet
    native: QTableLayer | QSpreadSheet

    @staticmethod
    def _normalize_data(data) -> pd.DataFrame:
        import pandas as pd

        if not isinstance(data, pd.DataFrame):
            try:
                mod = type(data).__module__.split(".", 1)[0]
            except AttributeError:
                mod = ""
            if mod == "polars":
                import polars as pl

                if isinstance(data, pl.DataFrame):
                    data = data.to_pandas()
                elif isinstance(data, pl.Series):
                    data = pd.DataFrame(data.to_pandas())
                else:
                    raise TypeError(f"{type(data)} cannot be converted.")
            elif mod == "xarray" and hasattr(data, "to_dataframe"):
                import xarray as xr

                if isinstance(data, xr.DataArray):
                    data = data.to_dataframe("val").reset_index()
                else:
                    raise TypeError(f"{type(data)} cannot be converted.")
            else:
                data = pd.DataFrame(data)
        return data

    def assign(self, other: dict[str, Any] = {}, /, **kwargs: dict[str, Any]) -> Self:
        """
        Assign new column(s) to the table.

        Examples
        --------
        >>> table.assign(a=[1, 2, 3], b=[2, 3, 4])
        >>> table.assign({"a": [1, 2, 3], "b": [2, 3, 4]})

        Returns
        -------
        TableBase
            Same table object with new columns.
        """
        import pandas as pd

        kwargs = dict(**other, **kwargs)
        if self._qwidget._data_raw.size == 0:
            # DataFrame.assign does not support updating empty DataFrame.
            self._qwidget.setDataFrame(pd.DataFrame(kwargs))
        else:
            serieses: dict[str, pd.Series] = {}
            for k, v in kwargs.items():
                serieses[str(k)] = pd.Series(v, index=self.data.index, name=k)

            self._qwidget.assignColumns(serieses)
        return self


@_doc.update_doc
class Table(_DataFrameTableLayer):
    """
    A table implemented with type checking.

    Parameters
    ----------
    data : DataFrame like, optional
        Table data to add.
    {name}{editable}{metadata}
    """

    _Default_Name = "table"
    _qwidget: QTableLayer
    native: QTableLayer

    def _create_backend(self, data: pd.DataFrame) -> QTableLayer:
        from tabulous._qt import QTableLayer

        return QTableLayer(data=data)


@_doc.update_doc
class SpreadSheet(_DataFrameTableLayer):
    """
    A table that behaves like a spreadsheet.

    Parameters
    ----------
    data : DataFrame like, optional
        Table data to add.
    {name}{editable}{metadata}
    """

    _qwidget: QSpreadSheet
    native: QSpreadSheet
    _Default_Name = "sheet"
    dtypes = _comp.ColumnDtypeInterface()

    def _create_backend(self, data: pd.DataFrame) -> QSpreadSheet:
        from tabulous._qt import QSpreadSheet

        return QSpreadSheet(data=data)

    def add_item_widget(self, row: int, column: int, widget):
        """Add a widget to a cell."""
        return self._qwidget._set_widget_at_index(row, column, widget)

    def _install_actions(self):
        from tabulous import commands as cmds

        _wrap = self._wrap_command
        # fmt: off
        _vheader_register = self.index.register
        _vheader_register("Insert row above", _wrap(cmds.selection.insert_row_above))
        _vheader_register("Insert row below", _wrap(cmds.selection.insert_row_below))
        _vheader_register("Remove selected rows", _wrap(cmds.selection.remove_selected_rows))  # noqa: E501
        self._qwidget._qtable_view.verticalHeader().addSeparator()

        _hheader_register = self.columns.register
        _hheader_register("Insert column left", _wrap(cmds.selection.insert_column_left))  # noqa: E501
        _hheader_register("Insert column right", _wrap(cmds.selection.insert_column_right))  # noqa: E501
        _hheader_register("Remove selected columns", _wrap(cmds.selection.remove_selected_columns))  # noqa: E501
        self._qwidget._qtable_view.horizontalHeader().addSeparator()
        _hheader_register("Column dtype", _wrap(cmds.selection.set_column_dtype))
        self._qwidget._qtable_view.horizontalHeader().addSeparator()

        _cell_register = self._qwidget.registerAction
        _cell_register("Insert/Remove > Insert a row above", _wrap(cmds.selection.insert_row_above))  # noqa: E501
        _cell_register("Insert/Remove > Insert a row below", _wrap(cmds.selection.insert_row_below))  # noqa: E501
        _cell_register("Insert/Remove > Remove rows", _wrap(cmds.selection.remove_selected_rows))  # noqa: E501
        self._qwidget.addSeparator()
        _cell_register("Insert/Remove > Insert a column on the left", _wrap(cmds.selection.insert_column_left))  # noqa: E501
        _cell_register("Insert/Remove > Insert a column on the right", _wrap(cmds.selection.insert_column_right))  # noqa: E501
        _cell_register("Insert/Remove > Remove columns", _wrap(cmds.selection.remove_selected_columns))  # noqa: E501
        self._qwidget.addSeparator()

        super()._install_actions()

        _cell_register("Cell widget > SpinBox", _wrap(cmds.selection.add_spinbox))
        _cell_register("Cell widget > Slider", _wrap(cmds.selection.add_slider))
        _cell_register("Cell widget > FloatSpinBox", _wrap(cmds.selection.add_float_spinbox))  # noqa: E501
        _cell_register("Cell widget > FloatSlider", _wrap(cmds.selection.add_float_slider))  # noqa: E501
        _cell_register("Cell widget > CheckBox", _wrap(cmds.selection.add_checkbox))
        _cell_register("Cell widget > RadioButton", _wrap(cmds.selection.add_radio_button))  # noqa: E501
        _cell_register("Cell widget > LineEdit", _wrap(cmds.selection.add_line_edit))
        self._qwidget.addSeparator("Cell widget ")
        _cell_register("Cell widget > Remove", _wrap(cmds.selection.remove_cell_widgets))  # noqa: E501

        # fmt: on
        return None


@_doc.update_doc
class GroupBy(TableBase):
    """
    A group of tables.

    Parameters
    ----------
    data : DataFrameGroupBy like, optional
        Groupby data to add.
    {name}{metadata}{update}
    """

    _Default_Name = "groupby"
    _qwidget: QTableGroupBy
    native: QTableGroupBy

    def _create_backend(self, data: pd.DataFrame) -> QTableGroupBy:
        from tabulous._qt import QTableGroupBy

        return QTableGroupBy(data=data)

    @staticmethod
    def _normalize_data(data):
        import pandas as pd
        from pandas.core.groupby.generic import DataFrameGroupBy

        if isinstance(data, DataFrameGroupBy):
            pass
        elif isinstance(data, (list, tuple, dict)):
            data_all = []
            group = "group"
            if isinstance(data, dict):
                it = data.items()
            else:
                it = enumerate(data)
            for key, df in it:
                df = pd.DataFrame(df, copy=True)
                if group in df.columns:
                    raise ValueError("Input data must not have a 'group' column.")
                data_all.append(df.assign(group=key))
            data = pd.concat(data_all, axis=0, ignore_index=True).groupby(group)
        else:
            raise TypeError("Cannot only add DataFrameGroupBy object.")
        return data

    @property
    def current_group(self):
        """Current group ID."""
        return self._qwidget.currentGroup()

    @current_group.setter
    def current_group(self, val) -> None:
        return self._qwidget.setCurrentGroup(val)


@_doc.update_doc
class TableDisplay(TableBase):
    """
    A table that is hotly reloaded by the given function.

    Parameters
    ----------
    data : callable, optional
        The loader function.
    {name}{metadata}{update}
    """

    _Default_Name = "display"
    _qwidget: QTableDisplay
    native: QTableDisplay

    def _create_backend(self, data: Callable[[], Any]) -> QTableDisplay:
        from tabulous._qt import QTableDisplay

        return QTableDisplay(loader=data)

    @staticmethod
    def _normalize_data(data):
        if not callable(data):
            raise TypeError("Can only add callable object.")
        return data

    @property
    def interval(self) -> int:
        """Interval of table refresh."""
        return self._qwidget.interval()

    @interval.setter
    def interval(self, value: int) -> None:
        """Set the interval of table refresh."""
        if not 10 <= value <= 10000:
            raise ValueError("Interval must be between 10 and 10000.")
        self._qwidget.setInterval(value)

    @property
    def loader(self) -> Callable[[], Any]:
        """Loader function."""
        return self._qwidget.loader()

    @loader.setter
    def loader(self, value: Callable[[], Any]) -> None:
        self._qwidget.setLoader(value)

    @property
    def running(self) -> bool:
        """True if the loading task is running"""
        return self._qwidget.running()

    @running.setter
    def running(self, value: bool) -> None:
        return self._qwidget.setRunning(value)


# deprecations


def foreground_colormap(self: TableBase, *args, **kwargs):
    """Deprecated method."""
    warnings.warn(
        "Method `table.foreground_colormap` is deprecated. "
        "Use `table.text_color.set` instead.",
        DeprecationWarning,
    )
    return self.text_color.set(*args, **kwargs)


def background_colormap(self: TableBase, *args, **kwargs):
    """Deprecated method."""
    warnings.warn(
        "Method `table.background_colormap` is deprecated. "
        "Use `table.background_color.set` instead.",
        DeprecationWarning,
    )
    return self.background_color.set(*args, **kwargs)


TableBase.foreground_colormap = foreground_colormap
TableBase.background_colormap = background_colormap



from __future__ import annotations
from typing import Any, Callable, Generic, TypeVar
from qtpy import QtWidgets as QtW, QtCore
from qtpy.QtWidgets import QAction

_T = TypeVar("_T")


class QActionRegistry(QtCore.QObject, Generic[_T]):
    """
    An contextmenu action registry.

    This class must be subclassed with other QWidget class like below.

    >>> class MyWidget(QWidget, QActionRegistry[int]):
    >>>     def __init__(self, parent=None):
    >>>         QWidget.__init__(self, parent)
    >>>         QActionRegistry.__init__(self)

    Now you can register any Python function with signature f(i: _T) to the context
    menu.
    """

    def __init__(self, *args, **kwargs) -> None:
        self._qt_context_menu = QContextMenu(self)

    def registerAction(self, location: str, f: Callable[[_T], Any]):
        """Register a function to the context menu at the given location."""
        locs = location.split(">")
        menu = self._qt_context_menu
        for loc in locs[:-1]:
            loc = loc.strip()
            a = menu.searchChild(loc)
            if a is None:
                menu = menu.addMenu(loc)
            elif not isinstance(a, QContextMenu):
                i = locs.index(loc)
                err_loc = ">".join(locs[:i])
                raise TypeError(f"{err_loc} is not a menu.")
            else:
                menu = a

        action = QAction(locs[-1], self)
        action.triggered.connect(lambda: f(self._qt_context_menu._current_index))
        menu.addAction(action)

        return f

    def unregisterAction(self, location: str):
        locs = location.split(">")
        menu = self._qt_context_menu
        for loc in locs[:-1]:
            loc = loc.strip()
            a = menu.searchChild(loc)
            if a is None:
                menu = menu.addMenu(loc)
            elif not isinstance(a, QContextMenu):
                i = locs.index(loc)
                err_loc = ">".join(locs[:i])
                raise TypeError(f"{err_loc} is not a menu.")
            else:
                menu = a
        action = menu._actions[loc[-1]]
        menu.removeAction(action)
        return None

    def addSeparator(self, location: str | None = None):
        """Add separator at the given location."""
        menu = self._qt_context_menu
        if location is None:
            locs = []
        else:
            locs = location.split(">")

        for loc in locs:
            loc = loc.strip()
            a = menu.searchChild(loc)
            if a is None:
                raise ValueError(f"{location} is not a valid location.")
            elif not isinstance(a, QContextMenu):
                i = locs.index(loc)
                err_loc = ">".join(locs[:i])
                raise TypeError(f"{err_loc} is not a menu.")
            else:
                menu = a

        menu.addSeparator()
        return None

    def execContextMenu(self, pos: QtCore.QPoint, index: _T) -> None:
        """Execute the context menu at the given index."""
        return self._qt_context_menu.execAtIndex(pos, index)


class QContextMenu(QtW.QMenu):
    """Contextmenu for the tabs on the QTabWidget."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self._current_index = None
        self._actions: dict[str, QAction | QContextMenu] = {}

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._actions})"

    def addMenu(self, title: str) -> QContextMenu:
        """Add a submenu to the contextmenu."""
        menu = self.__class__(self)
        action = super().addMenu(menu)
        action.setText(title)
        self._actions[title] = menu
        return menu

    def addAction(self, action: QAction) -> None:
        super().addAction(action)
        self._actions[action.text()] = action
        return None

    def execAtIndex(self, pos: QtCore.QPoint, index: _T):
        """Execute contextmenu at index."""
        if not self._actions:
            # don't show contextmenu if no action is registered.
            return None
        self._current_index = index
        try:
            self.exec_(pos)
        finally:
            self._current_index = None
        return None

    def searchChild(self, name: str) -> QAction | QContextMenu | None:
        """Return a action that matches the name."""
        return self._actions.get(name, None)



from ._table import QTableLayer, QSpreadSheet, QTableGroupBy, QTableDisplay
from ._table_stack import QTabbedTableStack
from ._mainwindow import QMainWindow, QMainWidget
from ._app import get_app

__all__ = [
    "QTableLayer",
    "QSpreadSheet",
    "QTableGroupBy",
    "QTableDisplay",
    "QTabbedTableStack",
    "QMainWindow",
    "QMainWidget",
    "get_app",
]



from __future__ import annotations
from typing import TYPE_CHECKING
from ._dialogs import choose_one

if TYPE_CHECKING:
    from tabulous.widgets._mainwindow import TableViewerBase

__all__ = ["open_table", "open_spreadsheet", "save_table", "open_sample"]

# fmt: off
SAMPLE_CHOICES = [
    "anagrams", "anscombe", "attention", "brain_networks", "car_crashes", "diamonds",
    "dots", "dowjones", "exercise", "flights", "fmri", "geyser", "glue", "healthexp",
    "iris", "mpg", "penguins", "planets", "seaice", "taxis", "tips", "titanic",
]
# fmt: on


def open_table(viewer: TableViewerBase):
    """Open a file as a table"""
    paths = viewer.history_manager.openFileDialog(mode="rm", caption="Open file(s)")
    for path in paths:
        viewer.open(path, type="table")
    return None


def open_spreadsheet(viewer: TableViewerBase):
    """Open a file as a spreadsheet"""
    paths = viewer.history_manager.openFileDialog(mode="rm", caption="Open file(s)")
    for path in paths:
        viewer.open(path, type="spreadsheet")
    return None


def save_table(viewer: TableViewerBase):
    """Save current table data"""
    if table := viewer.current_table:
        path = viewer.history_manager.openFileDialog(mode="w", caption="Save table")
        if path:
            table.save(path)
    return None


def save_table_to_source(viewer: TableViewerBase):
    """Save current table data to the source file if exists"""
    if table := viewer.current_table:
        if path := table.source.path:
            table.save(path)
        else:
            save_table(viewer)
    return None


def open_sample(viewer: TableViewerBase):
    """Open sample data"""
    out = choose_one(choice={"choices": SAMPLE_CHOICES, "nullable": False})
    if out is not None:
        viewer.open_sample(out)


def save_as_xlsx(viewer: TableViewerBase):
    """Save all tables to an Excel book"""
    path = viewer.history_manager.openFileDialog(
        mode="w", caption="Save table", filter="Excel book (*.xlsx; *.xls)"
    )
    if path:
        viewer.save_all(path)
    return None



from ._base import Component, TableComponent, ViewerComponent
from ._header import VerticalHeaderInterface, HorizontalHeaderInterface
from ._ranges import SelectionRanges, HighlightRanges, SelectedData
from ._cell import CellInterface
from ._plot import PlotInterface
from ._column_setting import (
    TextColormapInterface,
    BackgroundColormapInterface,
    TextFormatterInterface,
    ValidatorInterface,
    ColumnDtypeInterface,
)
from ._proxy import ProxyInterface
from ._viewer import Toolbar, Console, CommandPalette
from ._table_subset import TableSeries, TableSubset, TableLocIndexer, TableILocIndexer
from ._keymap import KeyMap

__all__ = [
    "Component",
    "TableComponent",
    "ViewerComponent",
    "VerticalHeaderInterface",
    "HorizontalHeaderInterface",
    "SelectionRanges",
    "HighlightRanges",
    "SelectedData",
    "CellInterface",
    "PlotInterface",
    "TextColormapInterface",
    "BackgroundColormapInterface",
    "TextFormatterInterface",
    "ValidatorInterface",
    "ColumnDtypeInterface",
    "ProxyInterface",
    "TableSeries",
    "TableSubset",
    "TableLocIndexer",
    "TableILocIndexer",
    "Toolbar",
    "Console",
    "CommandPalette",
    "KeyMap",
]



from __future__ import annotations

from abc import ABC, abstractproperty
from pathlib import Path
from types import MappingProxyType
import warnings
import weakref
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, Union
from psygnal import Signal, SignalGroup

from ._table import Table, SpreadSheet, GroupBy, TableDisplay
from ._tablelist import TableList
from ._sample import open_sample
from ._component import Toolbar, Console, CommandPalette
from ._source import Source
from . import _doc

from tabulous import _utils, _io
from tabulous.types import SelectionType, TabPosition, _TableLike, _SingleSelection
from tabulous.exceptions import UnreachableError
from tabulous.widgets._keymap_abc import SupportKeyMap

if TYPE_CHECKING:
    from tabulous.widgets._table import TableBase
    from tabulous._qt import QMainWindow, QMainWidget
    from tabulous._qt._dockwidget import QtDockWidget
    from tabulous._qt._mainwindow import _QtMainWidgetBase
    from tabulous._qt._mainwindow._namespace import Namespace
    from qtpy.QtWidgets import QWidget
    from magicgui.widgets import Widget
    import numpy as np
    import pandas as pd

PathLike = Union[str, Path, bytes]


class TableType(Enum):
    table = "table"
    spreadsheet = "spreadsheet"


class TableViewerSignal(SignalGroup):
    """Signal group for table viewer."""

    current_index = Signal(int)


class _AbstractViewer(ABC):
    @abstractproperty
    def current_table(self) -> TableBase | None:
        """Return the currently visible table."""

    @abstractproperty
    def current_index(self) -> int | None:
        """Return the index of currently visible table."""

    @abstractproperty
    def cell_namespace(self) -> Namespace:
        """Return the namespace of the cell editor."""

    @property
    def data(self) -> pd.DataFrame | None:
        """The data of the current table if exists."""
        table = self.current_table
        if table is None:
            return None
        return table.data

    @property
    def config(self) -> MappingProxyType:
        """Return the config info."""
        return _utils.get_config().as_immutable()

    def reset_choices(self, *_):
        pass

    def copy_data(
        self,
        selections: SelectionType | _SingleSelection | None = None,
        *,
        headers: bool = False,
        sep: str = "\t",
    ) -> None:
        """Copy selected cells to clipboard."""
        if selections is not None:
            self.current_table.selections = selections
        return self.current_table._qwidget.copyToClipboard(headers=headers, sep=sep)

    def paste_data(
        self,
        selections: SelectionType | _SingleSelection | None = None,
    ) -> None:
        """Paste from clipboard."""
        if selections is not None:
            self.current_table.selections = selections
        return self.current_table._qwidget.pasteFromClipBoard()


class TableViewerBase(_AbstractViewer, SupportKeyMap):
    """The base class of a table viewer widget."""

    events: TableViewerSignal
    _qwidget_class: type[_QtMainWidgetBase]

    toolbar = Toolbar()
    console = Console()
    command_palette = CommandPalette()

    def __init__(
        self,
        *,
        tab_position: TabPosition | str = TabPosition.top,
        show: bool = True,
    ):
        from tabulous._qt import get_app
        from tabulous._utils import get_post_initializers

        app = get_app()  # noqa: F841
        self._qwidget = self._qwidget_class(tab_position=tab_position)
        self._qwidget._table_viewer = self
        self._tablist = TableList(parent=self)
        self._link_events()

        self.events = TableViewerSignal()
        self._table_initializer = None
        _init = get_post_initializers()
        if _init is not None:
            viewer_initializer, self._table_initializer = _init
            viewer_initializer.initialize_viewer(self)

        if show:
            self.show(run=False)

    def __repr__(self) -> str:
        return f"<{type(self).__name__} widget at {hex(id(self))}>"

    @property
    def tables(self) -> TableList:
        """Return the table list object."""
        return self._tablist

    @property
    def current_table(self) -> TableBase | None:
        """Return the currently visible table."""
        if len(self.tables) > 0:
            return self.tables[self.current_index]
        return None

    @property
    def current_index(self) -> int | None:
        """Return the index of currently visible table."""
        stack = self._qwidget._tablestack
        if stack.isEmpty():
            return None
        return stack.currentIndex()

    @current_index.setter
    def current_index(self, index: int | str):
        if isinstance(index, str):
            index = self.tables.index(index)
        elif index < 0:
            index += len(self.tables)
        return self._qwidget._tablestack.setCurrentIndex(index)

    @property
    def native(self) -> _QtMainWidgetBase:
        """Return the native widget."""
        return self._qwidget

    @property
    def cell_namespace(self) -> Namespace:
        """Return the namespace of the cell editor."""
        return self._qwidget._namespace

    @property
    def history_manager(self):
        """The file I/O history manager."""
        return self._qwidget._hist_mgr

    @property
    def theme(self) -> str:
        """Get current widget theme name."""
        return self._qwidget._style_theme

    @theme.setter
    def theme(self, theme: str):
        self._qwidget.applyTheme(theme)

    def show(self, *, run: bool = True) -> None:
        """Show the widget."""
        self._qwidget.show()
        if run:
            from tabulous._qt._app import run_app

            run_app()
        return None

    def screenshot(self) -> np.ndarray:
        """Get screenshot of the widget."""
        return self._qwidget.screenshot()

    def save_screenshot(self, path: str):
        """Save screenshot of the widget."""
        from PIL import Image

        arr = self.screenshot()
        Image.fromarray(arr).save(path)
        return None

    @_doc.update_doc
    def add_table(
        self,
        data: _TableLike | None = None,
        *,
        name: str | None = None,
        editable: bool = False,
        copy: bool = True,
        metadata: dict[str, Any] | None = None,
        update: bool = False,
    ) -> Table:
        """
        Add data as a table.

        Parameters
        ----------
        data : DataFrame like, optional
            Table data to add.
        {name}{editable}{copy}{metadata}{update}

        Returns
        -------
        Table
            The added table object.
        """
        if copy:
            data = _copy_dataframe(data)
        table = Table(data, name=name, editable=editable, metadata=metadata)
        return self.add_layer(table, update=update)

    def add_spreadsheet(
        self,
        data: _TableLike | None = None,
        *,
        name: str | None = None,
        editable: bool = True,
        copy: bool = True,
        metadata: dict[str, Any] | None = None,
        update: bool = False,
        dtyped: bool = False,
    ) -> SpreadSheet:
        """
        Add data as a spreadsheet.

        Parameters
        ----------
        data : DataFrame like, optional
            Table data to add.
        {name}{editable}{copy}{metadata}{update}
        dtyped : bool, default is False
            If True, dtypes of the dataframe columns will be saved in the spreadsheet.
            Typed spreadsheet is safer for data recovery but less flexible for editing.

        Returns
        -------
        SpreadSheet
            The added spreadsheet object.
        """
        if copy:
            data = _copy_dataframe(data)
        table = SpreadSheet(data, name=name, editable=editable, metadata=metadata)
        if dtyped:
            table.dtypes.update(data.dtypes)
        return self.add_layer(table, update=update)

    def add_groupby(
        self,
        data,
        *,
        name: str | None = None,
        metadata: dict[str, Any] | None = None,
        update: bool = False,
    ) -> GroupBy:
        """
        Add a groupby object.

        Parameters
        ----------
        data : DataFrameGroupBy like object
            The groupby object to add.
        {name}{metadata}{update}

        Returns
        -------
        GroupBy
            A groupby table.
        """
        table = GroupBy(data, name=name, metadata=metadata)
        return self.add_layer(table, update=update)

    def add_loader(
        self,
        loader: Callable[[], _TableLike],
        *,
        name: str | None = None,
        metadata: dict[str, Any] | None = None,
        update: bool = False,
    ) -> TableDisplay:
        """
        Add a data frame loader function and continuously update the table.

        Parameters
        ----------
        loader : callable
            The loader function.
        {name}{metadata}{update}

        Returns
        -------
        TableDisplay
            A table display object.
        """
        table = TableDisplay(loader, name=name, metadata=metadata)
        return self.add_layer(table, update=update)

    def add_layer(self, input: TableBase, *, update: bool = False):
        if (
            update
            and (table := self.tables.get(input.name, None))
            and type(table) is type(input)
        ):
            table.data = input.data
            return table
        self.tables.append(input)
        self.current_index = -1  # activate the last table

        self._qwidget.setCellFocus()
        if self._table_initializer is not None:
            self._table_initializer.initialize_table(input)
        return input

    def open(self, path: PathLike, *, type: TableType | str = TableType.table) -> None:
        """
        Read a table data and add to the viewer.

        Parameters
        ----------
        path : path like
            File path.
        """
        path = Path(path)
        type = TableType(type)
        if type is TableType.table:
            fopen = self.add_table
        elif type is TableType.spreadsheet:
            fopen = self.add_spreadsheet
        else:
            raise UnreachableError(type)

        out = _io.open_file(path)
        if isinstance(out, dict):
            for sheet_name, df in out.items():
                table = fopen(df, name=sheet_name)
                table._source = Source(path)
        else:
            table = fopen(out, name=path.stem)
            table._source = Source(path)

        _utils.dump_file_open_path(path)
        return None

    def save(self, path: PathLike) -> None:
        """Save current table."""
        warnings.warn(
            "viewer.save() is deprecated. Use table.save() instead.",
            DeprecationWarning,
        )
        _io.save_file(path, self.current_table.data)
        return None

    def save_all(self, path: PathLike) -> None:
        """Save all tables."""
        path = Path(path)
        if path.is_dir():
            paths = [path / f"{table.name}.csv" for table in self.tables]
        elif path.name.count("*") == 1:
            paths = [path.replace("*", table.name) for table in self.tables]
        elif path.suffix in (".xlsx", ".xls"):
            import pandas as pd
            from tabulous._pd_index import is_ranged

            with pd.ExcelWriter(path, mode="w") as writer:
                for table in self.tables:
                    df = table.data
                    df.to_excel(
                        writer,
                        sheet_name=table.name,
                        index=not is_ranged(df.index),
                        header=not is_ranged(df.columns),
                    )
            return None
        else:
            raise ValueError("Invalid path.")

        for table, fp in zip(self.tables, paths):
            _io.save_file(fp, table.data)
        return None

    def open_sample(
        self,
        sample_name: str,
        *,
        plugin: str = "seaborn",
        type: TableType | str = TableType.table,
    ) -> Table:
        """Open a sample table."""
        df = open_sample(sample_name, plugin)
        type = TableType(type)
        if type is TableType.table:
            fopen = self.add_table
        elif type is TableType.spreadsheet:
            fopen = self.add_spreadsheet
        else:
            raise UnreachableError(type)
        return fopen(df, name=sample_name)

    def copy_data(
        self,
        selections: SelectionType | _SingleSelection | None = None,
        *,
        headers: bool = False,
        sep: str = "\t",
    ) -> None:
        """Copy selected cells to clipboard."""
        if selections is not None:
            self.current_table.selections = selections
        return self.current_table._qwidget.copyToClipboard(headers=headers, sep=sep)

    def paste_data(
        self,
        selections: SelectionType | _SingleSelection | None = None,
    ) -> None:
        """Paste from clipboard."""
        if selections is not None:
            self.current_table.selections = selections
        return self.current_table._qwidget.pasteFromClipBoard()

    def close(self):
        """Close the viewer."""
        return self._qwidget.close()

    @property
    def status(self) -> str:
        """Return the statup tip"""
        return self._status

    @status.setter
    def status(self, tip: str) -> None:
        """Set the status tip"""
        self._status = tip
        statusbar = self._qwidget.statusBar()
        return statusbar.showMessage(tip)

    def resize(self, width: int, height: int):
        """Resize the table viewer."""
        warnings.warn(
            "viewer.resize() is deprecated. Use `viewer.size` instead.",
            DeprecationWarning,
        )
        self.size = width, height

    @property
    def size(self) -> tuple[int, int]:
        """Return the size of the table viewer."""
        return self._qwidget.size().width(), self._qwidget.size().height()

    @size.setter
    def size(self, size: tuple[int, int]):
        """Set the size of the table viewer."""
        w, h = size
        return self._qwidget.resize(int(w), int(h))

    def _link_events(self):
        _tablist = self._tablist
        _qtablist = self._qwidget._tablestack

        @_tablist.events.inserted.connect
        def _insert_qtable(i: int):
            table = _tablist[i]
            _qtablist.addTable(table._qwidget, table.name)

        @_tablist.events.removed.connect
        def _remove_qtable(index: int, table: TableBase):
            _qtablist.blockSignals(True)
            try:
                _qtablist.takeTable(index)
            finally:
                _qtablist.blockSignals(False)

        @_tablist.events.moved.connect
        def _move_qtable(src: int, dst: int):
            # Evented list emits (0, 1) when move(0, 2) is called (v0.3.5 now).
            _qtablist.blockSignals(True)
            try:
                _qtablist.moveTable(src, dst)
            finally:
                _qtablist.blockSignals(False)

        @_tablist.events.renamed.connect
        def _rename_qtable(index: int, name: str):
            with _tablist.events.blocked():
                _qtablist.renameTable(index, name)

        @_qtablist.itemMoved.connect
        def _move_pytable(src: int, dst: int):
            """Move evented list when list is moved in GUI."""
            if src < dst:
                dst += 1
            with self._tablist.events.blocked():
                self._tablist.move(src, dst)

        @_qtablist.tableRenamed.connect
        def _rename_pytable(index: int, name: str):
            self._tablist.rename(index, name)

        @_qtablist.tableRemoved.connect
        def _remove_pytable(index: int):
            with self._tablist.events.blocked():
                del self._tablist[index]

        @_qtablist.tablePassed.connect
        def _pass_pytable(src, index: int, dst):
            src_ = _find_parent_table(src)
            dst_ = _find_parent_table(dst)
            dst_.tables.append(src_.tables.pop(index))

        _qtablist.itemDropped.connect(self.open)

        # reset choices when something changed in python table list
        _tablist.events.inserted.connect(self.reset_choices)
        _tablist.events.removed.connect(self.reset_choices)
        _tablist.events.moved.connect(self.reset_choices)
        _tablist.events.changed.connect(self.reset_choices)
        _tablist.events.renamed.connect(self.reset_choices)


class TableViewerWidget(TableViewerBase):
    """The non-main table viewer widget."""

    events: TableViewerSignal
    _qwidget: QMainWidget

    @property
    def _qwidget_class(self) -> QMainWidget:
        from tabulous._qt import QMainWidget

        return QMainWidget

    def add_widget(
        self,
        widget: Widget | QWidget,
        *,
        name: str = "",
    ):
        """Add a widget to the viewer as a child floating widget."""
        backend_widget, name = _normalize_widget(widget, name)
        backend_widget.setParent(self._qwidget, backend_widget.windowFlags())
        return backend_widget


class TableViewer(TableViewerBase):
    """The main table viewer widget."""

    events: TableViewerSignal
    _dock_widgets: weakref.WeakValueDictionary[str, QtDockWidget]
    _qwidget: QMainWindow

    @property
    def _qwidget_class(self) -> QMainWindow:
        from tabulous._qt import QMainWindow

        return QMainWindow

    def __init__(
        self,
        *,
        tab_position: TabPosition | str = TabPosition.top,
        show: bool = True,
    ):
        self._dock_widgets = weakref.WeakValueDictionary()
        self._status = ""
        super().__init__(
            tab_position=tab_position,
            show=show,
        )

    @property
    def config(self) -> MappingProxyType:
        """Return the config info."""
        from tabulous._utils import get_config

        return get_config().as_immutable()

    def add_dock_widget(
        self,
        widget: Widget | QWidget,
        *,
        name: str = "",
        area: str = "right",
        allowed_areas: list[str] | None = None,
    ):
        """
        Add a Qt or magic widget as a dock widget of the table viewer.

        Parameters
        ----------
        widget : magicgui.widgets.Widget or QWidget
            Widget to add to the table viewer.
        name : str, optional
            Name of the widget.
        area : str, default is "right"
            The initial area of the dock widget.
        allowed_areas : list of str, optional
            The allowed areas of the dock widget.
        """
        backend_widget, name = _normalize_widget(widget, name)

        dock = self._qwidget.addDockWidget(
            backend_widget, name=name, area=area, allowed_areas=allowed_areas
        )
        dock.setSourceObject(widget)
        self._dock_widgets[name] = dock
        dock.show()
        return dock

    def remove_dock_widget(self, name_or_widget: str | Widget | QWidget):
        """Remove dock widget from the table viewer."""
        if isinstance(name_or_widget, str):
            name = name_or_widget
            dock = self._dock_widgets[name_or_widget]
        else:
            for k, v in self._dock_widgets.items():
                if v is name_or_widget:
                    name = k
                    dock = v
                    break
            else:
                raise ValueError(f"Widget {name_or_widget!r} not found.")
        self._qwidget.removeDockWidget(dock)
        self._dock_widgets.pop(name)
        return None

    def reset_choices(self, *_):
        """Reset all the magicgui combo boxes."""
        for dock in self._dock_widgets.values():
            widget = dock.widget
            if hasattr(widget, "reset_choices"):
                widget.reset_choices()


class DummyViewer(_AbstractViewer):
    """
    The dummy object that emulates the table viewer.

    This is used when a table is used without a viewer, while make many commands
    still available from the table.
    """

    _namespace: Namespace | None = None

    def __init__(self, table: TableBase):
        self._table = table

    @property
    def current_table(self) -> TableBase:
        return self._table

    @property
    def current_index(self) -> int:
        return 0

    @property
    def cell_namespace(self) -> Namespace:
        """Return the namespace of the cell editor."""
        cls = self.__class__
        if cls._namespace is None:
            from tabulous._qt._mainwindow._namespace import Namespace
            from tabulous._utils import load_cell_namespace

            cls._namespace = Namespace()

            # update with user namespace
            cls._namespace.update_safely(load_cell_namespace())
        return cls._namespace


def _normalize_widget(widget: Widget | QWidget, name: str) -> tuple[QWidget, str]:
    if hasattr(widget, "native"):
        backend_widget = widget.native
        if not name:
            name = widget.name
    else:
        backend_widget = widget
        if not name:
            name = backend_widget.objectName()

    return backend_widget, name


def _copy_dataframe(data) -> pd.DataFrame:
    import pandas as pd

    if isinstance(data, pd.DataFrame):
        return data.copy()
    else:
        return Table._normalize_data(data)


def _find_parent_table(qwidget: _QtMainWidgetBase) -> TableViewerBase:
    x = qwidget
    while (parent := x.parent()) is not None:
        x = parent
        if hasattr(x, "_table_viewer"):
            return x._table_viewer
    raise RuntimeError



from __future__ import annotations
import re
from typing import Any, TYPE_CHECKING
from psygnal import Signal, SignalGroup, SignalInstance
from psygnal.containers import EventedList

from ._table import TableBase
from ._registry import SupportActionRegistration

if TYPE_CHECKING:
    from ._mainwindow import TableViewer

# Modified from psygnal/containers/_evented_list.py. See https://github.com/tlambert03/psygnal.
class NamedListEvents(SignalGroup):
    """Events available on EventedList.

    Attributes
    ----------
    inserting (index: int)
        emitted before an item is inserted at `index`
    inserted (index: int, value: Any)
        emitted after `value` is inserted at `index`
    removing (index: int)
        emitted before an item is removed at `index`
    removed (index: int, value: Any)
        emitted after `value` is removed at `index`
    moving (index: int, new_index: int)
        emitted before an item is moved from `index` to `new_index`
    moved (index: int, new_index: int, value: Any)
        emitted after `value` is moved from `index` to `new_index`
    changed (index: int or slice, old_value: Any or List[Any], value: Any or List[Any])
        emitted when `index` is set from `old_value` to `value`
    reordered (value: self)
        emitted when the list is reordered (eg. moved/reversed).
    child_event (index: int, object: Any, emitter: SignalInstance, args: tuple)
        emitted when an object in the list emits an event.
        Note that the EventedList must be created with `child_events=True` for this
        to be emitted.
    """

    inserting = Signal(int)  # idx
    inserted = Signal(int, object)  # (idx, value)
    removing = Signal(int)  # idx
    removed = Signal(int, object)  # (idx, value)
    moving = Signal(int, int)  # (src_idx, dest_idx)
    moved = Signal(tuple, object)  # ((src_idx, dest_idx), value)
    changed = Signal(object, object, object)  # (int | slice, old, new)
    reordered = Signal()
    child_event = Signal(int, object, SignalInstance, tuple)
    renamed = Signal(int, str)


class TableList(EventedList[TableBase], SupportActionRegistration["TableViewer", int]):
    events: NamedListEvents

    def __init__(self, parent: TableViewer):
        super().__init__()
        self.events = NamedListEvents()
        self._parent = parent
        self._install_contextmenu()

    @property
    def parent(self):
        """The parent viewer widget."""
        return self._parent

    @property
    def _qcontextmenu(self):
        """The QContextMenu widget."""
        return self.parent.native._tablestack._qt_context_menu

    def insert(self, index: int, table: TableBase):
        """Insert a table at index `index`."""
        if not isinstance(table, TableBase):
            raise TypeError(
                f"Cannot insert {type(table)} to {self.__class__.__name__}."
            )

        table.name = self._coerce_name(table.name, except_for=table)

        @table.events.renamed.connect
        def _renamed_signal(name: str):
            coerced_name = self._coerce_name(name, table)
            with table.events.renamed.blocked():
                table.name = coerced_name
            for i, tb in enumerate(self):
                if tb is table:
                    self.events.renamed.emit(i, coerced_name)
                    break
            return None

        super().insert(index, table)

    def index(self, value: TableBase | str, start: int = 0, stop: int = 999999) -> int:
        """Override of list.index(), also accepts str input."""
        if isinstance(value, str):
            for i, content in enumerate(self):
                if content.name == value:
                    return i
            else:
                raise ValueError(f"No table named {value}")
        else:
            return super().index(value, start, stop)

    def rename(self, index_or_name: int | str, name: str) -> None:
        """Rename a table name."""
        table = self[index_or_name]
        name = self._coerce_name(name, except_for=table)
        table.name = name
        return None

    def get(self, name: str, default: Any | None = None) -> TableBase | None:
        """Get a table with name `name` if exists."""
        for content in self:
            if content.name == name:
                return content
        else:
            return default

    def __getitem__(self, key):
        if isinstance(key, str):
            key = self.index(key)
        return super().__getitem__(key)

    def __delitem__(self, key):
        if isinstance(key, str):
            key = self.index(key)
        return super().__delitem__(key)

    def _coerce_name(self, name: str, except_for: TableBase):
        names = {content.name for content in self if content is not except_for}

        suffix = re.findall(r".*-(\d+)", name)
        if suffix:
            suf = suffix[0]
            new_name = name
            name = new_name.rstrip(suf)[:-1]
            i = int(suffix[0])
        else:
            new_name = name
            i = 0
        while new_name in names:
            new_name = f"{name}-{i}"
            i += 1

        return new_name

    def tile(self, indices: list[int], orientation: str = "horizontal") -> None:
        """Tile the tables in the list."""
        self._parent._qwidget._tablestack.tileTables(indices, orientation=orientation)
        return None

    def untile(self, indices: int | list[int]):
        """Untile the tables in the list."""
        if isinstance(indices, int):
            indices = [indices]
        for idx in indices:
            self._parent._qwidget._tablestack.untileTable(idx)
        return None

    def del_checked(self, index: int):
        table = self[index]
        if table._qwidget._edited and table.source.path is not None:
            self._parent._qwidget._tablestack.notifyNotSaved(index)
        else:
            del self[index]

    def _get_qregistry(self):
        return self._parent._qwidget._tablestack

    def _install_contextmenu(self):
        """Install the default contextmenu."""

        from tabulous import commands as cmds

        def _wrap(cmd):
            return lambda *_: cmd(self._parent)

        tablestack = self._parent._qwidget._tablestack

        # fmt: off
        self.register("Copy all", _wrap(cmds.table.copy_to_clipboard))
        self.register("Rename tab", _wrap(cmds.tab.rename_tab))
        self.register("Delete tab", _wrap(cmds.tab.delete_tab))
        tablestack.addSeparator()
        self.register("Reset filter/sort", _wrap(cmds.table.reset_proxy))
        tablestack.addSeparator()
        self.register("View > Horizontal dual view", _wrap(cmds.view.set_dual_h_mode))  # noqa: E501
        self.register("View > Vertical dual view", _wrap(cmds.view.set_dual_v_mode))  # noqa: E501
        self.register("View > Popup view", _wrap(cmds.view.set_popup_mode))  # noqa: E501
        self.register("View > Reset view", _wrap(cmds.view.reset_view_mode))  # noqa: E501
        self.register("Tile", _wrap(cmds.tab.tile_tables))
        self.register("Untile", _wrap(cmds.tab.untile_table))
        tablestack.addSeparator()
        # fmt: on



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Mapping,
    SupportsIndex,
    Tuple,
    Any,
    Iterator,
)
import warnings

from qtpy import QtGui
from qtpy.QtCore import Qt
from magicgui.widgets import Widget

from tabulous.exceptions import TableImmutableError
from tabulous.types import EvalInfo
from tabulous.color import ColorTuple
from tabulous._psygnal import InCellRangedSlot
from ._base import TableComponent
from tabulous.widgets._registry import SupportActionRegistration

if TYPE_CHECKING:
    import pandas as pd
    from tabulous.widgets import TableBase  # noqa: F401


class _Sequence2D(TableComponent):
    def __getitem__(self, key: tuple[int, int]):
        raise NotImplementedError()

    def _assert_integers(self, key: tuple[int, int]):
        r, c = key
        if not (isinstance(r, SupportsIndex) and isinstance(c, SupportsIndex)):
            raise TypeError("Cell label must be accessed by integer indices.")


class CellLabelInterface(_Sequence2D):
    def __getitem__(self, key: tuple[int, int]) -> str | None:
        """Get the label of a cell."""
        self._assert_integers(key)
        return self.parent.native.itemLabel(*key)

    def __setitem__(self, key: tuple[int, int], value: str):
        """Set the label of a cell."""
        self._assert_integers(key)
        return self.parent.native.setItemLabel(*key, value)

    def __delitem__(self, key: tuple[int, int]):
        """Delete the label of a cell."""
        self._assert_integers(key)
        return self.parent.native.setItemLabel(*key, None)


class CellReferenceInterface(
    TableComponent, Mapping[Tuple[int, int], InCellRangedSlot]
):
    """Interface to the cell references of a table."""

    def _table_map(self):
        return self.parent._qwidget._qtable_view._table_map

    def __getitem__(self, key: tuple[int, int]):
        return self._table_map()[key]

    def __iter__(self) -> Iterator[InCellRangedSlot]:
        return iter(self._table_map())

    def __len__(self) -> int:
        return len(self._table_map())

    def __repr__(self) -> str:
        slots = self._table_map()
        cname = type(self).__name__
        if len(slots) == 0:
            return f"{cname}()"
        s = ",\n\t".join(f"{k}: {slot!r}" for k, slot in slots.items())
        return f"{cname}(\n\t{s}\n)"


class CellBackgroundColorInterface(_Sequence2D):
    def __getitem__(self, key: tuple[int, int]) -> ColorTuple | None:
        """Get the background color of a cell."""
        self._assert_integers(key)
        model = self.parent.native.model()
        idx = model.index(*key)
        qcolor = model.data(idx, role=Qt.ItemDataRole.BackgroundRole)
        if isinstance(qcolor, QtGui.QColor):
            return ColorTuple(*qcolor.getRgb())
        return None


class CellForegroundColorInterface(_Sequence2D):
    def __getitem__(self, key: tuple[int, int]) -> ColorTuple | None:
        """Get the text color of a cell."""
        self._assert_integers(key)
        model = self.parent.native.model()
        idx = model.index(*key)
        qcolor = model.data(idx, role=Qt.ItemDataRole.TextColorRole)
        if isinstance(qcolor, QtGui.QColor):
            return ColorTuple(*qcolor.getRgb())
        return None


class CellDisplayedTextInterface(_Sequence2D):
    def __getitem__(self, key: tuple[int, int]) -> str:
        """Get the displayed text of a cell."""
        self._assert_integers(key)
        model = self.parent.native.model()
        idx = model.index(*key)
        return model.data(idx, role=Qt.ItemDataRole.DisplayRole)


class CellInterface(TableComponent, SupportActionRegistration["TableBase", int]):
    """
    Interface with table cells.

    This object can be used to emulate editing cells in the table.

    >>> table.cell[i, j]  # get the (i, j) cell value shown in the table
    >>> table.cell[i, j] = value  # set the (i, j) cell value
    >>> table.cell[i, j] = "&=np.mean(df.iloc[:, 0]"  # in-cell evaluation
    >>> del table.cell[i, j]  # delete the (i, j) cell value

    Label texts in the cell can be accessed by the ``label`` attribute.

    >>> table.cell.label[i, j]  # get the (i, j) cell label text
    >>> table.cell.label[i, j] = value  # set the (i, j) cell label text
    >>> del table.cell.label[i, j]  # delete the (i, j) cell label text

    Use ``register`` to register a contextmenu function.

    >>> @table.cell.register("My Action")
    >>> def my_action(table, index):
    ...     # do something
    """

    def __getitem__(self, key: tuple[int | slice, int | slice]):
        return self.parent._qwidget.model().df.iloc[key]

    def __setitem__(self, key: tuple[int | slice, int | slice], value: Any) -> None:
        table = self.parent
        if not table.editable:
            raise TableImmutableError("Table is not editable.")

        if isinstance(value, Widget):
            # add item widget
            r, c = key
            if isinstance(r, int) and isinstance(c, int):
                sheet = self._assert_spreadsheet()
                return sheet._qwidget._set_widget_at_index(*key, value)
            raise TypeError("Cannot set widget at slices.")

        import pandas as pd
        from tabulous._qt._table._base._line_edit import QCellLiteralEdit

        row, col = self._normalize_key(key)

        if isinstance(value, str) and QCellLiteralEdit._is_eval_like(value):
            if row.stop - row.start == 1 and col.stop - col.start == 1:
                expr, is_ref = QCellLiteralEdit._parse_ref(value)
                _r0, _c0 = row.start, col.start
                _r1 = table.proxy._get_proxy_object().get_source_index(_r0)
                info = EvalInfo(
                    pos=(_r0, _c0),
                    source_pos=(_r1, _c0),
                    expr=expr,
                    is_ref=is_ref,
                )
                table.events.evaluated.emit(info)
                return None
            else:
                raise ValueError("Cannot evaluate a multi-cell selection.")

        if isinstance(value, str) or not hasattr(value, "__iter__"):
            _value = [[value]]
        else:
            _value = value
        try:
            df = pd.DataFrame(_value)
        except ValueError:
            raise ValueError(f"Could not convert value {_value!r} to DataFrame.")

        if 1 in df.shape and (col.stop - col.start, row.stop - row.start) == df.shape:
            # it is natural to set an 1-D array without thinking of the direction.
            df = df.T

        table._qwidget.setDataFrameValue(row, col, df)
        return None

    def __delitem__(self, key: tuple[int | slice, int | slice]) -> None:
        """Deleting cell, equivalent to pushing Delete key."""

        table = self.parent
        if not table.editable:
            raise TableImmutableError("Table is not editable.")
        row, col = key
        table._qwidget.setSelections([(row, col)])
        table._qwidget.deleteValues()
        return None

    def _normalize_key(
        self,
        key: tuple[int | slice, int | slice],
    ) -> tuple[slice, slice]:
        if len(key) == 1:
            key = (key, slice(None))
        row, col = key

        if isinstance(row, slice):
            row = _normalize_slice(row, self.parent._qwidget.model().df.shape[0])
        else:
            row = slice(row, row + 1)

        if isinstance(col, slice):
            col = _normalize_slice(col, self.parent._qwidget.model().df.shape[1])
        else:
            col = slice(col, col + 1)
        return row, col

    @property
    def _qcontextmenu(self):
        """The QContextMenu widget."""
        return self.parent.native._qt_context_menu

    def _get_qregistry(self):
        return self.parent.native

    label = CellLabelInterface()
    ref = CellReferenceInterface()
    text_color = CellForegroundColorInterface()
    background_color = CellBackgroundColorInterface()
    text = CellDisplayedTextInterface()

    def selected_at(self, r: int, c: int) -> bool:
        """Check if a cell is selected."""
        for rr, cc in self.parent.selections:
            if rr.start <= r < rr.stop and cc.start <= c < cc.stop:
                return True
        return False

    def get_label(self, r: int, c: int) -> str | None:
        """Get the label of a cell."""
        warnings.warn(
            "get_label is deprecated. Use `table.cell.label[r, c]` instead.",
            DeprecationWarning,
        )
        return self.label[r, c]

    def set_label(self, r: int, c: int, text: str):
        """Set the label of a cell."""
        warnings.warn(
            f"set_label is deprecated. Use `table.cell.label[r, c] = {text!r}` "
            "instead.",
            DeprecationWarning,
        )
        self.label[r, c] = text

    def set_labeled_data(
        self,
        r: int,
        c: int,
        data: dict[str, Any] | pd.Series | tuple,
        sep: str = "",
    ):
        """
        Set given data with cell labels.

        Parameters
        ----------
        r : int
            Starting row index.
        c : int
            Starting column index.
        data : dict[str, Any], pd.Series or (named) tuple
            Data to be set.
        sep : str, optional
            Separator added at the end of labels.
        """
        import pandas as pd

        ndata = len(data)
        if isinstance(data, dict):
            _data = pd.Series(data)
        elif isinstance(data, tuple):
            index = getattr(data, "_fields", range(ndata))
            _data = pd.Series(data, index=index)
        elif isinstance(data, pd.Series):
            _data = data.copy()
        else:
            raise TypeError(f"Cannot convert {data!r} to Series.")

        if sep:
            _data.index = [f"{idx}{sep}" for idx in _data.index]

        self.parent._qwidget.setLabeledData(slice(r, r + ndata), slice(c, c + 1), _data)
        return None


def _normalize_slice(sl: slice, size: int) -> slice:
    start = sl.start
    stop = sl.stop

    if sl.step not in (None, 1):
        raise ValueError("Row slice step must be 1.")

    if start is None:
        start = 0
    elif start < 0:
        start = size + start
    if stop is None:
        stop = size
    elif stop < 0:
        stop = size + stop
    return slice(start, stop)



from __future__ import annotations
from abc import abstractmethod
from typing import (
    Hashable,
    TYPE_CHECKING,
    TypeVar,
    Any,
    Union,
    MutableMapping,
    Iterator,
    Callable,
    Mapping,
    Sequence,
)
from functools import wraps
import warnings

import numpy as np

from tabulous.types import ColorMapping, ColorType
from tabulous.color import InvertedColormap, OpacityColormap, BrightenedColormap
from tabulous._dtype import get_converter, get_converter_from_type, isna
from tabulous._colormap import segment_by_float, segment_by_time
from ._base import Component, TableComponent

if TYPE_CHECKING:
    from pandas.core.dtypes.dtypes import ExtensionDtype
    from tabulous.widgets._table import TableBase, SpreadSheet  # noqa: F401

    _DtypeLike = Union[ExtensionDtype, np.dtype]

    from typing_extensions import TypeGuard, Self
    import pandas as pd

    _Formatter = Union[Callable[[Any], str], str, None]
    _Validator = Callable[[Any], None]
    _TimeType = Union[pd.Timestamp, pd.Timedelta]
    _NumberLike = Union[int, float, _TimeType]
    _Interpolatable = Union[
        Mapping[_NumberLike, ColorType],
        Sequence[tuple[_NumberLike, ColorType]],
        Sequence[ColorType],
    ]

T = TypeVar("T")
_F = TypeVar("_F", bound=Callable)
_Void = object()


class _DictPropertyInterface(TableComponent, MutableMapping[str, _F]):
    @abstractmethod
    def _get_dict(self) -> dict[str, _F]:
        """Get dict of colormaps."""

    @abstractmethod
    def _set_value(self, key: str, func: _F):
        """Set colormap at key."""

    def __getitem__(self, key: str) -> _F:
        return self._get_dict()[key]

    def __setitem__(self, key: str, func: _F) -> None:
        self.set(key, func)
        return None

    def __delitem__(self, key: str) -> None:
        return self._set_value(key, None)

    def __repr__(self) -> str:
        clsname = type(self).__name__
        _dict = self._get_dict()
        if _dict:
            _args = ",\n\t".join(f"{k!r}: {v}" for k, v in _dict.items())
            return f"{clsname}(\n\t{_args}\n)"
        return f"{clsname}()"

    def __len__(self) -> str:
        return len(self._get_dict())

    def __iter__(self) -> Iterator[str]:
        return iter(self._get_dict())

    def set(self, column_name: str, func: _F = _Void):
        """Set function to given column name."""

        def _wrapper(f: _F) -> _F:
            self._set_value(column_name, f)
            return f

        return _wrapper(func) if func is not _Void else _wrapper

    def reset(self, column_name) -> Self:
        """Reset function for the given column name."""
        self._set_value(column_name, None)
        return self

    def __call__(self, *args, **kwargs):
        # backwards compatibility
        return self.set(*args, **kwargs)


class _ColormapInterface(_DictPropertyInterface[ColorMapping]):
    """Abstract interface to the column colormap."""

    def set(
        self,
        column_name: str,
        colormap: ColorMapping = _Void,
        *,
        interp_from: _Interpolatable | None = None,
        infer_parser: bool = True,
        opacity: float | None = None,
    ):
        """
        Set colormap for the given column.

        Parameters
        ----------
        column_name : str
            Name of the column.
        colormap : ColorMapping, optional
            Colormap function or colormap name.
        interp_from : list/dict of (value, color) or two colors, optional
            Create colormap by interpolating given colors. If list or dict
            of colors are given, the colors are interpolated linearly between
            each adjacent value set. If two colors are given, the colors are
            interpolated linearly between the minimum and maximum values.
        infer_parser : bool, default is True
            If true, infer the parser from the column dtype and use it before
            the values are passed to the colormap function.
        opacity : float, optional
            If given, apply opacity to the colormap.
        """

        def _wrapper(f: ColorMapping) -> ColorMapping:
            if callable(f):
                if infer_parser:
                    parser = self._get_converter(f, column_name)
                    _f = wraps(f)(lambda x: f(parser(x)))
                else:
                    _f = f
                if opacity is not None:
                    _f = OpacityColormap.from_colormap(_f, opacity)
            else:
                # void or None
                _f = f
            self._set_value(column_name, _f)
            return f

        if isinstance(colormap, Mapping):
            return _wrapper(lambda x: colormap.get(x, None))
        elif colormap is _Void:
            if interp_from is None:
                return _wrapper
            else:
                return self.set(
                    column_name,
                    self._from_interpolatable(interp_from, column_name),
                    infer_parser=False,
                    opacity=opacity,
                )
        elif isinstance(colormap, str):
            colormap = self._get_mpl_colormap(column_name, colormap)
            return _wrapper(colormap)
        else:
            return _wrapper(colormap)

    def _from_interpolatable(
        self, seq: _Interpolatable, column_name: str
    ) -> ColorMapping:

        ds = self.parent.native._get_sub_frame(column_name)
        kind = ds.dtype.kind

        if isinstance(seq, Mapping):
            _seq = list(seq.items())
        elif not isinstance(seq[0], tuple) or len(seq[0]) != 2:
            # not list[tuple[number, ColorType]]
            vmin, vmax = ds.min(), ds.max()
            if kind not in "mM":
                _seq = zip(np.linspace(vmin, vmax, len(seq)), seq)
            else:
                import pandas as pd

                if kind == "m":
                    _seq = zip(pd.timedelta_range(vmin, vmax, periods=len(seq)), seq)
                else:
                    _seq = zip(pd.date_range(vmin, vmax, periods=len(seq)), seq)
        else:
            _seq = seq

        if kind in "mM":
            return segment_by_time(_seq, kind)
        else:
            return segment_by_float(_seq)

    def _get_mpl_colormap(self, column_name: str, colormap: str) -> ColorMapping:
        from matplotlib.cm import get_cmap

        mpl_cmap = get_cmap(colormap)

        def _cmap(x):
            return np.asarray(mpl_cmap(int(x * 255))) * 255

        return self._simple_cmap_for_column(column_name, _cmap)

    def _simple_cmap_for_column(self, column_name: str, cmap: ColorMapping):
        """
        Create a colormap for a column, with min/max as the color limits.

        Parameters
        ----------
        column_name : str
            Name of the column.
        cmap : [0, 1] -> ColorType
            Colormap function
        """
        ds = self.parent.native._get_sub_frame(column_name)
        vmin, vmax = ds.min(), ds.max()
        if ds.dtype.kind in "uif":

            def _cmap(x) -> ColorType:
                x = float(x)
                if isna(x):
                    return None
                ratio = (x - vmin) / (vmax - vmin)
                ratio = max(0.0, min(1.0, ratio))
                return cmap(ratio)

        elif ds.dtype.kind in "mM":
            vmin: _TimeType
            vmax: _TimeType
            vmin, vmax = vmin.value, vmax.value
            converter = get_converter(ds.dtype)

            def _cmap(x):
                x = converter(x).value
                ratio = (x - vmin) / (vmax - vmin)
                ratio = max(0.0, min(1.0, ratio))
                return cmap(ratio)

        elif ds.dtype.kind == "b":
            _cmap = cmap
        else:
            raise TypeError(f"Cannot infer colormap for dtype {ds.dtype}")
        return _cmap

    def _get_converter(self, f: ColorMapping, column_name: str):
        table = self.parent
        parser = None

        # try to infer parser from function annotations
        _ann = f.__annotations__
        if (key := next(iter(_ann.keys()), None)) and key != "return":
            arg_type = _ann[key]
            parser = get_converter_from_type(arg_type)

        elif _is_spreadsheet(table) and (dtype := table.dtypes.get(column_name, None)):
            # try to infer parser from table column dtype
            parser = get_converter(dtype)

        else:
            dtype = table.data[column_name].dtype
            parser = get_converter(dtype)

        return parser

    def invert(self, column_name: str):
        """Invert the colormap for a column."""
        self.set(
            column_name,
            InvertedColormap.from_colormap(self[column_name]),
            infer_parser=False,
        )
        return None

    def set_opacity(self, column_name: str, opacity: float):
        """Set the opacity value for a column."""
        self.set(
            column_name,
            OpacityColormap.from_colormap(self[column_name], opacity),
            infer_parser=False,
        )
        return None

    def adjust_brightness(self, column_name: str, factor: float):
        """
        Adjust the brightness of a column.

        Parameters
        ----------
        column_name : str
            Name of the column to adjust.
        factor : float
            Brightening factor. -1.0 converts any color to black, while 1.0
            to white.
        """
        self.set(
            column_name,
            BrightenedColormap.from_colormap(self[column_name], factor),
            infer_parser=False,
        )
        return None


class TextColormapInterface(_ColormapInterface):
    """Interface to the column text colormap."""

    def _get_dict(self) -> dict[str, ColorMapping]:
        return self.parent._qwidget.model()._foreground_colormap

    def _set_value(self, key: str, cmap: ColorMapping):
        return self.parent.native.setForegroundColormap(key, cmap)


class BackgroundColormapInterface(_ColormapInterface):
    """Interface to the column background colormap."""

    def _get_dict(self) -> dict[str, ColorMapping]:
        return self.parent._qwidget.model()._background_colormap

    def _set_value(self, key: str, cmap: ColorMapping):
        return self.parent.native.setBackgroundColormap(key, cmap)


class TextFormatterInterface(_DictPropertyInterface["_Formatter"]):
    """Interface to the column background colormap."""

    def _get_dict(self) -> dict[str, _Formatter]:
        return self.parent._qwidget.model()._text_formatter

    def _set_value(self, key: str, cmap: _Formatter):
        return self.parent.native.setTextFormatter(key, cmap)


class ValidatorInterface(_DictPropertyInterface["_Validator"]):
    """Interface to the column validator."""

    def _get_dict(self) -> dict[str, _Validator]:
        return self.parent._qwidget.model()._text_formatter

    def _set_value(self, key: str, validator: _Validator):
        return self.parent.native.setDataValidator(key, validator)


def _is_spreadsheet(table: TableBase) -> TypeGuard[SpreadSheet]:
    return table.table_type == "SpreadSheet"


class ColumnDtypeInterface(Component["SpreadSheet"], MutableMapping[str, "_DtypeLike"]):
    """Interface to the column dtype of spreadsheet."""

    def _get_dtype_map(self):
        return self.parent._qwidget._columns_dtype

    def __getitem__(self, key: str) -> _DtypeLike | None:
        """Get the dtype of the given column name."""
        return self._get_dtype_map().get(key, None)

    def __setitem__(self, key: str, dtype: Any) -> None:
        """Set a dtype to the given column name."""
        return self.parent._qwidget.setColumnDtype(key, dtype)

    def __delitem__(self, key: str) -> None:
        """Reset the dtype to the given column name."""
        return self.parent._qwidget.setColumnDtype(key, None)

    def __repr__(self) -> str:
        clsname = type(self).__name__
        _args = ",\n\t".join(f"{k!r}: {v}" for k, v in self._get_dtype_map().items())
        return f"{clsname}(\n\t{_args}\n)"

    def __len__(self) -> str:
        return len(self._get_dtype_map())

    def __iter__(self) -> Iterator[Hashable]:
        return iter(self._get_dtype_map())

    def set(
        self,
        name: str,
        dtype: Any,
        *,
        validation: bool = True,
        formatting: bool = True,
    ) -> None:
        """Set dtype and optionally default validator and formatter."""
        self.parent._qwidget.setColumnDtype(name, dtype)
        if validation:
            self.parent._qwidget._set_default_data_validator(name)
        if formatting:
            self.parent._qwidget._set_default_text_formatter(name)
        return None

    def set_dtype(self, *args, **kwargs) -> None:
        """Deprecated alias for set()."""
        warnings.warn(
            "set_dtype() is deprecated, use set() instead.",
            DeprecationWarning,
        )
        return self.set(*args, **kwargs)



from __future__ import annotations

from pathlib import Path
import logging
import ast
from typing import TYPE_CHECKING, Sequence
from qtpy import QtWidgets as QtW, QtCore, QtGui
from qtpy.QtCore import Qt, Signal

from tabulous._qt._toolbar._toolbutton import QColoredToolButton
from tabulous._sort_filter_proxy import (
    FilterType,
    FilterInfo,
    ComposableFilter,
    ComposableSorter,
)
from superqt import QEnumComboBox

from tabulous.exceptions import UnreachableError

if TYPE_CHECKING:
    from typing_extensions import Self
    from tabulous._qt._table import QBaseTable
    import pandas as pd

ICON_DIR = Path(__file__).parent / "_icons"
logger = logging.getLogger("tabulous")


class HeaderAnchorMixin:
    @classmethod
    def install_to_table(cls, table: QBaseTable, index: int, *args, **kwargs) -> Self:
        """Install this widget to table at index."""
        with table._mgr.merging():
            self = cls(*args, **kwargs)
            table.setHorizontalHeaderWidget(index, self)
        return self

    def on_installed(self, table: QBaseTable, index: int):
        pass

    def on_uninstalled(self, table: QBaseTable, index: int):
        pass


class _QHeaderSectionButton(QColoredToolButton, HeaderAnchorMixin):
    def __init__(self, parent: QtW.QWidget = None):
        super().__init__(parent)
        self.setFixedSize(16, 16)
        self.setCursor(Qt.CursorShape.PointingHandCursor)
        eff = QtW.QGraphicsOpacityEffect()
        self._opacity = 0.3
        eff.setOpacity(self._opacity)
        self.setGraphicsEffect(eff)
        self._effect = eff

    def enterEvent(self, event: QtCore.QEvent) -> None:
        self._effect.setOpacity(1.0)

    def leaveEvent(self, event: QtCore.QEvent) -> None:
        self._effect.setOpacity(self._opacity)

    def updateColorByBackground(self, bg: QtGui.QColor):
        whiteness = bg.red() + bg.green() + bg.blue()
        self._white_background = whiteness > 128 * 3
        if self._white_background:
            self.updateColor("#1E1E1E")
        else:
            self.updateColor("#CCCCCC")


class QHeaderSortButton(_QHeaderSectionButton):
    sortSignal = Signal(bool)
    resetSignal = Signal()

    def __init__(self, parent: QtW.QWidget = None, ascending: bool = True):
        super().__init__(parent)

        self.setIcon(ICON_DIR / "sort_table.svg")
        self._ascending = ascending
        self.clicked.connect(self._toggle)
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self._show_menu)
        self.setToolTip("Sort by this column (right click to reset)")

    def _toggle(self):
        self._ascending = not self._ascending
        self.sortSignal.emit(self._ascending)

    def _show_menu(self, pos):
        menu = QtW.QMenu(self)
        menu.addAction("Reset", lambda: self.resetSignal.emit())
        menu.exec_(self.mapToGlobal(pos))

    def ascending(self) -> bool:
        return self._ascending

    @classmethod
    def install_to_table(cls, table: QBaseTable, index: int, ascending: bool = True):
        with table._mgr.merging():
            self = cls(ascending=ascending)
            table.setHorizontalHeaderWidget(index, self)
        return self

    def on_installed(self, table: QBaseTable, index: int):
        logger.debug(f"Installing sort button at index {index}")

        def _sort():
            f = table._proxy._obj
            if isinstance(f, ComposableSorter):
                table._set_proxy(f.switch())
            elif f is None:
                table._set_proxy(ComposableSorter({index}, True))
            else:
                raise RuntimeError("Sort function is not a ComposableSorter.")

        def _reset():
            if isinstance(f, ComposableSorter):
                with table._mgr.merging(
                    lambda cmds: f"Remove filter button at index {index}"
                ):
                    table.setHorizontalHeaderWidget(index, None)
                    table._set_proxy(f.decompose(index))
            else:
                raise RuntimeError("Sort function is not a ComposableSorter.")

        f = table._proxy._obj
        if not isinstance(f, ComposableSorter):
            f = ComposableSorter({index}, self.ascending())
        table._set_proxy(f.compose(index))
        self.sortSignal.connect(_sort)
        self.resetSignal.connect(_reset)
        if _viewer := table.parentViewer():
            self.updateColorByBackground(_viewer.backgroundColor())
        return None

    def on_uninstalled(self, table: QBaseTable, index: int):
        logger.debug(f"Uninstalling sort button at index {index}")
        self.sortSignal.disconnect()
        f = table._proxy._obj
        if isinstance(f, ComposableSorter):
            table._set_proxy(f.decompose(index))
        return None


class QHeaderFilterButton(_QHeaderSectionButton):
    def __init__(self, parent: QtW.QWidget | None = None):
        super().__init__(parent)
        self.setPopupMode(QtW.QToolButton.ToolButtonPopupMode.InstantPopup)
        self.setIcon(ICON_DIR / "filter.svg")
        self.setToolTip("Filter by this column")

    def on_installed(self, table: QBaseTable, index: int):
        logger.debug(f"Installing filter at index {index}")

        def _filter(info: FilterInfo):
            f = table._proxy._obj
            if isinstance(f, ComposableFilter):
                table._set_proxy(f.compose(index, info))
            elif f is None:
                table._set_proxy(ComposableFilter({index: info}))
            else:
                raise RuntimeError("Current proxy is not a ComposableFilter.")
            # TODO: this is incompatible with undo/redo
            # self._opacity = 0.3 if info.type is FilterType.none else 0.7
            # self._effect.setOpacity(self._opacity)
            return None

        def _reset():
            f = table._proxy._obj
            if isinstance(f, ComposableFilter):
                with table._mgr.merging(
                    lambda cmds: f"Remove filter button at index {index}"
                ):
                    table.setHorizontalHeaderWidget(index, None)
                    table._set_proxy(f.decompose(index))
            else:
                raise RuntimeError("Sort function is not a ComposableFilter.")

        f = table._proxy._obj
        if not isinstance(f, ComposableFilter):
            table._set_proxy(ComposableFilter())
        column = table.model().df.columns[index]
        menu = _QFilterMenu(table._get_sub_frame(column), parent=self)
        self.setMenu(menu)
        menu._filter_widget.called.connect(_filter)
        menu._filter_widget.reset.connect(_reset)
        if _viewer := table.parentViewer():
            self.updateColorByBackground(_viewer.backgroundColor())
        return None

    def on_uninstalled(self, table: QBaseTable, index: int):
        logger.debug(f"Uninstalling filter at index {index}")
        f = table._proxy._obj
        if isinstance(f, ComposableFilter):
            table._set_proxy(f.decompose(index))

        return None


class _QFilterMenu(QtW.QMenu):
    def __init__(self, ds: pd.Series, parent: QtW.QWidget = None):
        super().__init__(parent)
        self._ds = ds
        action = QtW.QWidgetAction(self)
        self._filter_widget = _QFilterWidget(ds)
        self._filter_widget.called.connect(self.hide)
        self._filter_widget.reset.connect(self.hide)
        action.setDefaultWidget(self._filter_widget)
        self.addAction(action)
        self._filter_widget.requireResize.connect(self.resize)

    def showEvent(self, a0: QtGui.QShowEvent) -> None:
        self._filter_widget._cbox.setFocus()
        return super().showEvent(a0)


class _QFilterWidget(QtW.QWidget):
    called = Signal(FilterInfo)
    reset = Signal()
    requireResize = Signal(QtCore.QSize)

    def __init__(self, ds: pd.Series, parent: QtW.QWidget = None):
        super().__init__(parent)
        self.setMinimumWidth(150)
        self._ds = ds
        self._cbox = QEnumComboBox(enum_class=FilterType)
        self._cbox.setMinimumWidth(100)
        self._cbox.setFont(QtGui.QFont("Arial", 10))
        self._value_edit = QtW.QLineEdit()
        self._string_edit = QtW.QLineEdit()
        self._value_edit.setFixedWidth(84)
        self._string_edit.setFixedWidth(84)
        self._unique_select = QMultiCheckBoxes()
        self._call_button = QtW.QPushButton("Apply")
        self._call_button.setToolTip("Apply filter to the column")
        self._reset_button = QtW.QPushButton("Reset")
        self._reset_button.setToolTip("Reset the filter of the column")
        self._setup_ui()

        self._cbox.currentEnumChanged.connect(self._type_changed)
        self._call_button.clicked.connect(self._call_button_clicked)
        self._reset_button.clicked.connect(lambda: self.reset.emit())
        self._type_changed(FilterType.none)

    def _type_changed(self, val: FilterType):
        self._value_edit.setVisible(False)
        self._string_edit.setVisible(False)
        self._unique_select.setVisible(False)
        if val.requires_list:
            self._unique_select.setVisible(True)
            self._unique_select.setChoices(self.fetch_unique())
        elif val.requires_number:
            self._value_edit.setVisible(True)
            self._value_edit.setFocus()
        elif val.requires_text:
            self._string_edit.setVisible(True)
            self._string_edit.setFocus()
        elif val is not FilterType.none:
            raise UnreachableError(f"Unreachable: {val}")
        self.requireResize.emit(self.sizeHint())

    def _call_button_clicked(self):
        return self.called.emit(self.get_filter_info())

    def get_filter_info(self) -> FilterInfo:
        ftype: FilterType = self._cbox.currentEnum()
        if ftype.requires_number:
            arg = ast.literal_eval(self._value_edit.text())
        elif ftype.requires_text:
            arg = self._string_edit.text()
        elif ftype.requires_list:
            arg = self._unique_select.value()
        else:
            arg = None
        return FilterInfo(ftype, arg)

    def fetch_unique(self):
        unique = self._ds.unique()
        if len(unique) > 108:
            raise ValueError("Too many unique values")
        return unique

    def _setup_ui(self):
        _layout = QtW.QVBoxLayout()
        _layout.setContentsMargins(2, 2, 2, 2)

        _layout.addWidget(QtW.QLabel("Filter by:"))

        _middle = QtW.QWidget()
        _middle_layout = QtW.QHBoxLayout()
        _middle_layout.setContentsMargins(0, 0, 0, 0)
        _middle_layout.addWidget(self._cbox, alignment=Qt.AlignmentFlag.AlignLeft)
        _middle_layout.addWidget(
            self._value_edit, alignment=Qt.AlignmentFlag.AlignRight
        )
        _middle_layout.addWidget(
            self._string_edit, alignment=Qt.AlignmentFlag.AlignRight
        )
        _middle.setLayout(_middle_layout)

        _layout.addWidget(_middle)
        _layout.addWidget(self._unique_select)

        _bottom = QtW.QWidget()
        _bottom_layout = QtW.QHBoxLayout()
        _bottom_layout.setContentsMargins(0, 0, 0, 0)
        _bottom_layout.addWidget(self._call_button)
        _bottom_layout.addWidget(self._reset_button)
        _bottom.setLayout(_bottom_layout)
        _layout.addWidget(_bottom)
        self.setLayout(_layout)


class QMultiCheckBoxes(QtW.QListWidget):
    def __init__(self, parent: QtW.QWidget = None):
        super().__init__(parent)
        self.setSelectionMode(QtW.QAbstractItemView.SelectionMode.NoSelection)
        self.horizontalScrollBar().setVisible(False)
        self.setFixedHeight(120)
        self._choices = []

    def setChoices(self, choices: Sequence):
        self.clear()
        self._choices = choices
        for c in choices:
            text = repr(c)
            item = QtW.QListWidgetItem()
            self.addItem(item)
            checkbox = QtW.QCheckBox(text)
            checkbox.setChecked(False)
            self.setItemWidget(item, checkbox)
        return None

    def iter_items(self):
        for i in range(self.count()):
            yield self.item(i)

    def value(self) -> list:
        return [
            self._choices[i]
            for i in range(self.count())
            if self.itemWidget(self.item(i)).isChecked()
        ]

    if TYPE_CHECKING:

        def itemWidget(self, item: QtW.QListWidgetItem) -> QtW.QCheckBox:
            ...



from __future__ import annotations
from typing import Any, Callable, TypeVar
import warnings
import inspect

from qtpy import QtWidgets as QtW, QtGui
from magicgui import magicgui
from magicgui.widgets import Widget, Dialog
from tabulous.widgets import MagicTable
from tabulous.types import TableData

from ._register import find_table_viewer_ancestor

_F = TypeVar("_F", bound=Callable)


def dialog_factory(function: _F) -> _F:
    from magicgui.widgets import create_widget

    def _runner(parent=None, **param_options):
        # create a list of widgets, similar to magic_signature
        widgets: list[Widget] = []
        callbacks: dict[str, Callable[[Widget], Any]] = {}
        sig = inspect.signature(function)
        for name, param in sig.parameters.items():
            opt = param_options.get(name, {})
            opt.setdefault("gui_only", False)
            changed_cb = opt.pop("changed", None)
            if param.default is not inspect.Parameter.empty:
                opt.setdefault("value", param.default)
            if param.annotation is not inspect.Parameter.empty:
                opt.setdefault("annotation", param.annotation)
            opt.setdefault("name", name)
            kwargs: dict[str, Any] = {}
            for k in "value", "annotation", "name", "label", "widget_type", "gui_only":
                if k in opt:
                    kwargs[k] = opt.pop(k)
            widget = create_widget(**kwargs, options=opt)
            if changed_cb is not None:
                assert callable(changed_cb)
                callbacks[name] = changed_cb
            widgets.append(widget)

        dlg = Dialog(widgets=widgets)

        for name, cb in callbacks.items():
            dlg[name].changed.connect(lambda: cb(dlg))

        # if return annotation "TableData" is given, add a preview widget.
        if function.__annotations__.get("return") is TableData:
            table = MagicTable(
                data=[],
                name="preview",
                editable=False,
                tooltip="Preview of the result using the head of the input data.",
                gui_only=True,
            )
            table.zoom = 0.8
            dlg.append(table)

            @dlg.changed.connect
            def _on_value_change(*_):
                import pandas as pd

                kwargs = dlg.asdict()
                # Check the first data frame is not too large.
                argname, val = next(iter(kwargs.items()))
                if isinstance(val, pd.DataFrame):
                    num = 8400
                    if val.size > num:
                        kwargs[argname] = val.head(num // val.shape[1])
                try:
                    table.data = function(**kwargs)
                except Exception:
                    table.data = []

            dlg.changed.emit()

        dlg.native.setParent(parent, dlg.native.windowFlags())
        dlg._shortcut = QtW.QShortcut(QtGui.QKeySequence("Ctrl+W"), dlg.native)
        dlg._shortcut.activated.connect(dlg.close)
        dlg.reset_choices()
        if dlg.exec():
            out = function(**dlg.asdict())
        else:
            out = None
        return out

    return _runner


def dialog_factory_mpl(function: _F) -> _F:
    def _runner(parent=None, **param_options):
        dlg = magicgui(function, **param_options)

        from tabulous._qt._plot import QtMplPlotCanvas

        style = None
        bg = None
        if parent is not None:
            if viewer := find_table_viewer_ancestor(parent):
                if not viewer._qwidget._white_background:
                    style = "dark_background"
                bg = viewer._qwidget.backgroundColor().name()

        plt = QtMplPlotCanvas(style=style)
        if bg:
            plt.set_background_color(bg)

        dlg.native.layout().addWidget(plt)
        dlg.height = 400
        dlg.width = 280

        @dlg.changed.connect
        def _on_value_change(*_):
            kwargs = dlg.asdict()
            kwargs["ax"] = plt.ax
            if kwargs.get("ref", False):
                kwargs["ref"] = False
            try:
                plt.cla()
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    function(**kwargs)
                plt.draw()
            except Exception:
                pass

        dlg.changed.emit()

        dlg.native.setParent(parent, dlg.native.windowFlags())
        dlg._shortcut = QtW.QShortcut(QtGui.QKeySequence("Ctrl+W"), dlg.native)
        dlg._shortcut.activated.connect(dlg.close)
        dlg.called.connect(lambda: dlg.close())
        return dlg.show()

    return _runner



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Iterable, Iterator, Sequence, SupportsIndex


class TableAnchorBase(ABC):
    @abstractmethod
    def insert_rows(self, row: int, count: int) -> None:
        ...

    @abstractmethod
    def insert_columns(self, col: int, count: int) -> None:
        ...

    @abstractmethod
    def remove_rows(self, row: int, count: int) -> None:
        ...

    @abstractmethod
    def remove_columns(self, col: int, count: int) -> None:
        ...


class RectRange(TableAnchorBase):
    def __init__(
        self,
        rsl: slice = slice(0, 0),
        csl: slice = slice(0, 0),
    ):
        assert type(rsl) is slice and type(csl) is slice
        self._rsl = rsl
        self._csl = csl

    @classmethod
    def new(cls, r: slice | SupportsIndex, c: slice | SupportsIndex):
        if isinstance(r, SupportsIndex):
            r = slice(r, r + 1)
        if isinstance(c, SupportsIndex):
            c = slice(c, c + 1)
        return cls(r, c)

    @classmethod
    def from_shape(cls, shape: tuple[int, int]):
        return cls(slice(0, shape[0]), slice(0, shape[1]))

    def __repr__(self):
        return f"RectRange[{_fmt_slice(self._rsl)}, {_fmt_slice(self._csl)}]"

    def __contains__(self, other: tuple[int, int]):
        r, c = other
        rsl = self._rsl
        csl = self._csl
        rlower = rsl.start <= r if rsl.start is not None else True
        rupper = r < rsl.stop if rsl.stop is not None else True
        clower = csl.start <= c if csl.start is not None else True
        cupper = c < csl.stop if csl.stop is not None else True
        return rlower and rupper and clower and cupper

    def includes(self, other: RectRange) -> bool:
        if isinstance(other, MultiRectRange):
            return all(self.includes(rng) for rng in other)

        r0_s, r1_s = self._rsl.start, self._rsl.stop
        c0_s, c1_s = self._csl.start, self._csl.stop
        r0_o, r1_o = other._rsl.start, other._rsl.stop
        c0_o, c1_o = other._csl.start, other._csl.stop

        return (
            _le(r0_s, r0_o) and _ge(r1_s, r1_o) and _le(c0_s, c0_o) and _ge(c1_s, c1_o)
        )

    def overlaps_with(self, other: RectRange) -> bool:
        if isinstance(other, MultiRectRange):
            return other.overlaps_with(self)
        r0_s, r1_s = self._rsl.start, self._rsl.stop
        c0_s, c1_s = self._csl.start, self._csl.stop
        r0_o, r1_o = other._rsl.start, other._rsl.stop
        c0_o, c1_o = other._csl.start, other._csl.stop

        return _overlap_1d(r0_s, r1_s, r0_o, r1_o) and _overlap_1d(
            c0_s, c1_s, c0_o, c1_o
        )

    def insert_rows(self, row: int, count: int) -> None:
        """Insert rows and update slices in-place."""
        self._rsl = translate_slice(self._rsl, row, count)
        return None

    def insert_columns(self, col: int, count: int) -> None:
        """Insert columns and update slices in-place."""
        self._csl = translate_slice(self._csl, col, count)
        return None

    def remove_rows(self, row: int, count: int):
        """Remove rows and update slices in-place."""
        self._rsl = translate_slice(self._rsl, row, -count)
        return None

    def remove_columns(self, col: int, count: int):
        """Remove columns and update slices in-place."""
        self._csl = translate_slice(self._csl, col, -count)
        return None

    def is_empty(self) -> bool:
        """True if the range is empty."""
        r0_s, r1_s = self._rsl.start, self._rsl.stop
        c0_s, c1_s = self._csl.start, self._csl.stop
        if r0_s is None or r1_s is None or c0_s is None or c1_s is None:
            return False
        return self._rsl.start >= self._rsl.stop or self._csl.start >= self._csl.stop

    def iter_ranges(self) -> Iterator[tuple[slice, slice]]:
        return iter([(self._rsl, self._csl)])

    def as_iloc(self) -> tuple[slice, slice]:
        return self._rsl, self._csl

    def as_iloc_string(self, df_expr: str = "df") -> str:
        return f"{df_expr}.iloc[{_parse_slice(self._rsl)}, {_parse_slice(self._csl)}]"


def _DO_NOTHING(*args, **kwargs):
    return None


class AnyRange(RectRange):
    """Contains any indices."""

    def __init__(self):
        super().__init__(slice(None), slice(None))

    def __contains__(self, item) -> bool:
        return True

    def __repr__(self):
        return "AnyRange[...]"

    def includes(self, other: RectRange) -> bool:
        return True

    def overlaps_with(self, other: RectRange) -> bool:
        return True

    insert_rows = _DO_NOTHING
    insert_columns = _DO_NOTHING
    remove_rows = _DO_NOTHING
    remove_columns = _DO_NOTHING

    def is_empty(self) -> bool:
        return False


class NoRange(RectRange):
    """Contains no index."""

    def __init__(self):
        super().__init__(slice(0, -1), slice(0, -1))

    def __contains__(self, item) -> bool:
        return False

    def __repr__(self):
        return "NoRange[...]"

    def includes(self, other: RectRange) -> bool:
        return False

    def overlap_with(self, other: RectRange) -> bool:
        return False

    insert_rows = _DO_NOTHING
    insert_columns = _DO_NOTHING
    remove_rows = _DO_NOTHING
    remove_columns = _DO_NOTHING

    def is_empty(self) -> bool:
        return True


class MultiRectRange(RectRange):
    """Range composed of multiple RectRanges."""

    def __init__(self, ranges: Sequence[RectRange]):
        self._ranges = ranges

    @classmethod
    def from_slices(cls, slices: Iterable[tuple[slice, slice]]):
        """Construct from list of slices."""
        return cls([RectRange(rsl, csl) for rsl, csl in slices])

    def __repr__(self):
        s = ", ".join(repr(rng) for rng in self)
        return f"MultiRectRange[{s}]"

    def __contains__(self, other: tuple[int, int]):
        return any(other in rng for rng in self)

    def __iter__(self) -> Iterator[RectRange]:
        return iter(self._ranges)

    def includes(self, other: RectRange) -> bool:
        if isinstance(other, MultiRectRange):
            return all(self.includes(rng) for rng in other)
        return any(rng.includes(other) for rng in self)

    def overlaps_with(self, other: RectRange) -> bool:
        if isinstance(other, MultiRectRange):
            return any(self.overlaps_with(rng) for rng in other)
        return any(rng.overlaps_with(other) for rng in self)

    def insert_rows(self, row: int, count: int) -> None:
        """Insert rows and update slices in-place."""
        for rng in self:
            rng.insert_rows(row, count)

    def insert_columns(self, col: int, count: int) -> None:
        """Insert columns and update slices in-place."""
        for rng in self:
            rng.insert_columns(col, count)

    def remove_rows(self, row: int, count: int):
        """Remove rows and update slices in-place."""
        for rng in self:
            rng.remove_rows(row, count)

    def remove_columns(self, col: int, count: int):
        """Remove columns and update slices in-place."""
        for rng in self:
            rng.remove_columns(col, count)

    def is_empty(self) -> bool:
        """True if ANY of the range is empty."""
        return any(rng.is_empty() for rng in self)

    def iter_ranges(self) -> Iterator[tuple[slice, slice]]:
        for rng in self:
            yield from rng.iter_ranges()

    def as_iloc(self):
        raise TypeError("Cannot convert MultiRectRange to iloc")


def _fmt_slice(sl: slice) -> str:
    s0 = sl.start if sl.start is not None else ""
    s1 = sl.stop if sl.stop is not None else ""
    return f"{s0}:{s1}"


def _parse_slice(sl: slice) -> str:
    """Convert slice to 'a:b' representation or to int if possible"""
    if sl.start is not None and sl.stop is not None:
        s0 = sl.start
        s1 = sl.stop
        if s0 == s1 - 1:
            return str(int(s0))
    s0 = sl.start if sl.start is not None else ""
    s1 = sl.stop if sl.stop is not None else ""
    return f"{s0}:{s1}"


def _le(r0_s, r0_o):
    if r0_s is not None:
        if r0_o is None:
            return False
        else:
            return r0_s <= r0_o
    else:
        return True


def _ge(r1_s, r1_o):
    if r1_s is not None:
        if r1_o is None:
            return False
        else:
            return r1_o <= r1_s
    else:
        return True


INF = float("inf")


def _overlap_1d(
    r0_s: int | None, r1_s: int | None, r0_o: int | None, r1_o: int | None
) -> bool:
    if r0_s is None:
        r0_s = 0
    if r1_s is None:
        r1_s = INF
    if r0_o is None:
        r0_o = 0
    if r1_o is None:
        r1_o = INF

    return r0_o < r1_s and r0_s < r1_o


def translate_slice(sl: slice, index: int, count: int) -> slice:
    """
    Return the translated slice suppose row/column changed at given index.

    Parameters
    ----------
    sl : slice
        The slice to translate.
    index : int
        The starting index of the row/column that changed.
    count : int
        The number of rows/columns that changed.
    """
    start, stop = sl.start, sl.stop

    if count > 0:  # insertion
        if start is not None and start >= index:
            start = start + count
        if stop is not None and stop > index:
            stop = stop + count
    else:  # deletion
        if start is not None and start > index:
            start = start + count
        if stop is not None and stop > index:
            stop = stop + count

    return slice(start, stop)



from __future__ import annotations
import weakref
from typing import (
    Generic,
    Literal,
    TYPE_CHECKING,
    TypeVar,
    overload,
    Any,
)

if TYPE_CHECKING:
    from typing_extensions import Self
    from tabulous.widgets._table import TableBase, SpreadSheet  # noqa: F401
    from tabulous.widgets._mainwindow import TableViewerBase  # noqa: F401

T = TypeVar("T")


class _NoRef:
    """No reference."""


class Component(Generic[T]):
    _no_ref = _NoRef()

    def __init__(self, parent: T | _NoRef = _no_ref):
        if parent is self._no_ref:
            self._instances: dict[int, Self] = {}
        else:
            self._instances = None
        self._parent_ref = weakref.ref(parent)

    @property
    def parent(self) -> T:
        """The parent object of this component."""
        out = self._parent_ref()
        if out is None:
            raise ReferenceError("Parent has been garbage collected.")
        return out

    def __repr__(self) -> str:
        return f"<{type(self).__name__} of {self.parent!r}>"

    @overload
    def __get__(self, obj: Literal[None], owner=None) -> Self[_NoRef]:
        ...

    @overload
    def __get__(self, obj: Any, owner=None) -> Self[T]:
        ...

    def __get__(self, obj, owner=None):
        if obj is None:
            return self
        if isinstance(obj, Component):
            obj = obj.parent
        _id = id(obj)
        if (out := self._instances.get(_id)) is None:
            out = self._instances[_id] = self.__class__(obj)
        return out

    def __set__(self, obj: Any, value: Any) -> None:
        if obj is None:
            raise AttributeError("Cannot set attribute.")
        _id = id(obj)
        if (ins := self._instances.get(_id)) is None:
            ins = self._instances[_id] = self.__class__(obj)

        return ins._set_value(value)

    def _set_value(self, value: Any):
        raise AttributeError("Cannot set attribute.")


class TableComponent(Component["TableBase"]):
    def _assert_spreadsheet(self) -> SpreadSheet:
        sheet = self.parent
        if sheet.table_type != "SpreadSheet":
            raise TypeError(
                f"{sheet.table_type!r} does not support insert. Use "
                "SpreadSheet instead."
            )
        return sheet


class ViewerComponent(Component["TableViewerBase"]):
    def show(self):
        """Show the component."""
        self.visible = True

    def hide(self):
        """Hide the component."""
        self.visible = False

    @property
    def visible(self) -> bool:
        """Visibility of the component."""
        raise NotImplementedError



from __future__ import annotations
from typing import Callable, Any
from enum import Enum, auto
from qtpy import QtWidgets as QtW
from qtpy.QtCore import Qt
import numpy as np
import pandas as pd

from tabulous.widgets import Table
from tabulous._dtype import get_dtype, isna
from tabulous._magicgui import ToggleSwitches

__all__ = ["exec_formatter_dialog"]


def _format_float(value, ndigits: int = 4) -> str:
    """convert string to int or float if possible"""
    if isna(value):
        text = "nan"
    elif 0.1 <= abs(value) < 10 ** (ndigits + 1) or value == 0:
        text = f"{value:.{ndigits}f}"
    else:
        text = f"{value:.{ndigits-1}e}"

    return text


def _format_int(value, ndigits: int = 4) -> str:
    if 0.1 <= abs(value) < 10 ** (ndigits + 1) or value == 0:
        text = str(value)
    else:
        text = f"{value:.{ndigits-1}e}"

    return text


def _format_complex(value: complex, ndigits: int = 3) -> str:
    if isna(value):
        text = "nan"
    elif 0.1 <= abs(value) < 10 ** (ndigits + 1) or value == 0:
        text = f"{value.real:.{ndigits}f}{value.imag:+.{ndigits}f}j"
    else:
        text = f"{value.real:.{ndigits-1}e}{value.imag:+.{ndigits-1}e}j"

    return text


_DEFAULT_FORMATTERS: dict[str, Callable[[Any], str]] = {
    "u": _format_int,
    "i": _format_int,
    "f": _format_float,
    "c": _format_complex,
}


class DefaultFormatter:
    """
    The default formatter function.

    This class is a simple wrapper around a callable. Spreadsheet needs to know if
    a formatter is the default one when dtype is changed.
    """

    def __init__(self, dtype: Any):
        self._dtype = get_dtype(dtype)
        self._formatter = _DEFAULT_FORMATTERS.get(self._dtype.kind, str)

    def __call__(self, value: Any) -> None:
        return self._formatter(value)

    def __repr__(self) -> str:
        return f"DefaultFormatter[{self._dtype.name}]"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, DefaultFormatter):
            return False
        return self._dtype == other._dtype


class NumberFormat(Enum):
    """Formats used for integers."""

    default = auto()
    decimal = auto()
    exponential = auto()


class FloatFormat(Enum):
    """Formats used for floating numbers."""

    default = auto()
    decimal = auto()
    exponential = auto()
    percent = auto()


class ComplexFormat(Enum):
    """Formats used for complex numbers."""

    default = auto()
    decimal = auto()
    exponential = auto()
    polar = auto()


class TimedeltaFormat(Enum):
    """Formats used for timedelta."""

    default = auto()
    day = auto()
    hour = auto()
    minute = auto()
    second = auto()
    hour_min_second = auto()
    min_second = auto()
    hour_min = auto()


class QDigitsSpinBox(QtW.QSpinBox):
    def __init__(self, value: int = 4):
        super().__init__()
        self.setMinimum(0)
        self.setMaximum(12)
        self.setValue(value)
        self.setSuffix(" digits")
        self.setToolTip("Number of digits to display.")


class _QFormatterDialog(QtW.QDialog):
    def __init__(self, ds: pd.Series, parent: QtW.QWidget | None = None) -> None:
        super().__init__(parent)
        _layout = QtW.QHBoxLayout()
        self._name = ds.name
        self._preview_table = Table(ds, editable=False)
        self._preview_table.native.setMaximumWidth(120)
        self._left_panel = QtW.QWidget()
        self._left_panel.setLayout(QtW.QVBoxLayout())
        self._left_panel.layout().setContentsMargins(0, 0, 0, 0)

        _layout.addWidget(self._left_panel)
        _layout.addWidget(titled("Preview", self._preview_table.native))
        self._init()
        self.setLayout(_layout)
        btnbox = QtW.QDialogButtonBox(
            QtW.QDialogButtonBox.StandardButton.Ok
            | QtW.QDialogButtonBox.StandardButton.Cancel,
        )
        btnbox.accepted.connect(self.accept)
        btnbox.rejected.connect(self.reject)
        self.addWidget(btnbox)
        self.resize(450, 320)

    def _init(self):
        """Initialize UI."""

    def addWidget(self, widget: QtW.QWidget):
        self._left_panel.layout().addWidget(widget)
        return None

    def toFormatter(self, val=None) -> str:
        """Convert widget paramters to a formatter text."""
        raise NotImplementedError()


class _QNumberFormatterDialog(_QFormatterDialog):
    _ENUM = NumberFormat

    def _init(self) -> None:
        btns = ToggleSwitches(choices=self._ENUM, value=self._ENUM.default)
        self._btns = btns
        self.addWidget(btns.native)

        self._ndigits = QDigitsSpinBox()
        self.addWidget(self._ndigits)

        btns.changed.connect(self._on_choice_changed)
        self._ndigits.valueChanged.connect(lambda: self._on_choice_changed())
        self._on_choice_changed()

    def _on_choice_changed(self, val=None):
        raise NotImplementedError()


class QIntFormatterDialog(_QNumberFormatterDialog):
    def _on_choice_changed(self, val=None):
        if val is None:
            val = self._btns.value
        val = self._ENUM(val)
        self._ndigits.setVisible(val == self._ENUM.exponential)
        self._preview_table.text_formatter(self._name, self.toFormatter(val))
        return self._preview_table.refresh()

    def toFormatter(self, val=None) -> str:
        """Convert widget paramters to a formatter text."""
        if val is None:
            val = self._btns.value

        n = self._ndigits.value()
        if val == self._ENUM.default:
            fmt = None
        elif val == self._ENUM.decimal:
            fmt = "{}"
        elif val == self._ENUM.exponential:
            fmt = f"{{:.{n}g}}"
        else:
            raise RuntimeError()
        return fmt


class QFloatFormatterDialog(_QNumberFormatterDialog):
    _ENUM = FloatFormat

    def _on_choice_changed(self, val=None):
        if val is None:
            val = self._btns.value
        val = self._ENUM(val)
        self._ndigits.setVisible(val != self._ENUM.default)
        self._preview_table.text_formatter(self._name, self.toFormatter(val))
        return self._preview_table.refresh()

    def toFormatter(self, val=None) -> str:
        """Convert widget paramters to a formatter text."""
        if val is None:
            val = self._btns.value

        n = self._ndigits.value()
        if val == self._ENUM.default:
            fmt = None
        elif val == self._ENUM.decimal:

            def fmt(x):
                return f"{float(x):.{n}f}"

        elif val == self._ENUM.exponential:

            def fmt(x):
                return f"{float(x):.{n}g}"

        elif val == self._ENUM.percent:

            def fmt(x):
                return f"{float(x):.{n}%}"

        else:
            raise RuntimeError()
        return fmt


class QComplexFormatterDialog(_QNumberFormatterDialog):
    _ENUM = ComplexFormat

    def _init(self) -> None:
        super()._init()
        self._iunit = QtW.QLineEdit("j")
        self._iunit.setToolTip("Imaginary unit.")
        self._iunit.textChanged.connect(lambda: self._on_choice_changed())
        self.addWidget(self._iunit)

    def _on_choice_changed(self, val=None):
        if val is None:
            val = self._btns.value
        val = self._ENUM(val)
        self._ndigits.setVisible(val != self._ENUM.default)
        self._preview_table.text_formatter(self._name, self.toFormatter(val))
        return self._preview_table.refresh()

    def toFormatter(self, val=None) -> None | Callable[[complex], str]:
        """Convert widget paramters to a formatter text."""
        if val is None:
            val = self._btns.value
        iunit = self._iunit.text()
        n = self._ndigits.value()
        if val == self._ENUM.default:
            return None
        elif val == self._ENUM.decimal:
            return lambda x: f"{x.real:.{n}f}{x.imag:+.{n}f}{iunit}"
        elif val == self._ENUM.exponential:
            return lambda x: f"{x.real:.{n}e}{x.imag:+.{n}e}{iunit}"
        elif val == self._ENUM.polar:
            return lambda x: f"{abs(x):.{n}f}e^({np.angle(x):+.{n}f}{iunit})"
        else:
            raise RuntimeError(val)


class QBoolFormatterDialog(_QFormatterDialog):
    def _init(self) -> None:
        self._true = QtW.QLineEdit("True")
        self._false = QtW.QLineEdit("False")
        self._true.textChanged.connect(self._on_text_changed)
        self._false.textChanged.connect(self._on_text_changed)
        self.addWidget(labeled("True:", self._true, 50))
        self.addWidget(labeled("False:", self._false, 50))

    def _on_text_changed(self, val=None):
        return self._preview_table.text_formatter(self._name, self.toFormatter(val))

    def toFormatter(self, val=None) -> Callable[[bool], str]:
        """Convert widget paramters to a formatter text."""
        return lambda x: self._true.text() if x else self._false.text()


class QDateTimeFormatterDialog(_QFormatterDialog):
    def _init(self) -> None:
        label = QtW.QLabel(
            "%Y: year\n"
            "%m: month\n"
            "%d: day\n"
            "%H: hour\n"
            "%M: minute\n"
            "%S: second\n"
        )
        self._fmt = QtW.QLineEdit("%Y-%m-%d")
        self.addWidget(label)
        self.addWidget(labeled("Format:", self._fmt))
        self._fmt.textChanged.connect(self._on_text_changed)

    def _on_text_changed(self, val=None):
        return self._preview_table.text_formatter(self._name, self.toFormatter(val))

    def toFormatter(self, val: str = None) -> Callable[[pd.Timestamp], str]:
        """Convert widget paramters to a formatter text."""
        if val is None:
            val = self._fmt.text()

        return lambda ts: ts.strftime(val)


class QTimeDeltaFormatterDialog(_QFormatterDialog):
    def _init(self) -> None:
        self._btns = ToggleSwitches(
            choices=TimedeltaFormat, value=TimedeltaFormat.default
        )
        self.addWidget(self._btns.native)

        self._ndigits = QDigitsSpinBox(0)
        self.addWidget(self._ndigits)

        self._btns.changed.connect(lambda: self._on_choice_changed())
        self._ndigits.valueChanged.connect(lambda: self._on_choice_changed())

    def _on_choice_changed(self, val=None):
        return self._preview_table.text_formatter(self._name, self.toFormatter(val))

    def toFormatter(self, val: str = None) -> Callable[[pd.Timedelta], str]:
        """Convert widget paramters to a formatter text."""
        if val is None:
            val = self._btns.value
        val = TimedeltaFormat(val)
        n = self._ndigits.value()

        if val == TimedeltaFormat.default:
            return lambda td: str(td)
        elif val == TimedeltaFormat.day:
            return lambda td: f"{td.days} days"
        elif val == TimedeltaFormat.hour:
            return lambda td: f"{td.total_seconds() / 3600:.{n}f} h"
        elif val == TimedeltaFormat.minute:
            return lambda td: f"{td.total_seconds() / 60:.{n}f} min"
        elif val == TimedeltaFormat.second:
            return lambda td: f"{td.total_seconds():.{n}f} s"
        elif val == TimedeltaFormat.hour_min_second:
            return lambda td: _to_hms(td, n)
        elif val == TimedeltaFormat.hour_min:
            return lambda td: _to_hm(td, n)
        elif val == TimedeltaFormat.min_second:
            return lambda td: _to_ms(td, n)
        else:
            raise RuntimeError(val)


def _to_hms(td: pd.Timedelta, n: int) -> str:
    """Convert timedelta to hours, minutes, seconds."""
    total_sec = td.total_seconds()
    hours, res_min = divmod(total_sec, 3600)
    minutes, seconds = divmod(res_min, 60)
    if seconds < 10:
        return f"{int(hours):0>2}:{int(minutes):0>2}:0{seconds:.{n}f}"
    return f"{int(hours):0>2}:{int(minutes):0>2}:{seconds:.{n}f}"


def _to_hm(td: pd.Timedelta, n: int) -> str:
    """Convert timedelta to hours, minutes, seconds."""
    total_min = td.total_seconds() / 60
    hours, minutes = divmod(total_min, 60)
    if minutes < 10:
        return f"{int(hours):0>2}:0{minutes:.{n}f}"
    return f"{int(hours):0>2}:{minutes:.{n}f}"


def _to_ms(td: pd.Timedelta, n: int) -> str:
    """Convert timedelta to hours, minutes, seconds."""
    total_sec = td.total_seconds()
    minutes, seconds = divmod(total_sec, 60)
    if seconds < 10:
        return f"{int(minutes):0>2}:0{seconds:.{n}f}"
    return f"{int(minutes):0>2}:{seconds:.{n}f}"


class _LabeledWidget(QtW.QWidget):
    def __init__(
        self,
        label: str,
        widget: QtW.QWidget,
        width: int | None = None,
        orientation: Qt.Orientation = Qt.Orientation.Horizontal,
    ):
        super().__init__()
        self._label = QtW.QLabel(label)
        self._widget = widget
        if orientation == Qt.Orientation.Horizontal:
            layout = QtW.QHBoxLayout()
            layout.setAlignment(
                Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter
            )
        elif orientation == Qt.Orientation.Vertical:
            layout = QtW.QVBoxLayout()
            layout.setAlignment(
                Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignHCenter
            )
            self._label.setAlignment(Qt.AlignmentFlag.AlignHCenter)
        layout.setContentsMargins(0, 0, 0, 0)

        layout.addWidget(self._label)
        layout.addWidget(self._widget)
        self.setLayout(layout)

        if width is not None and orientation == Qt.Orientation.Horizontal:
            self._label.setFixedWidth(width)


def labeled(label: str, widget: QtW.QWidget, width: int | None = None) -> QtW.QWidget:
    """Create a widget with a label."""
    return _LabeledWidget(label, widget, width)


def titled(title: str, widget: QtW.QWidget) -> QtW.QWidget:
    """Create a widget with a title."""
    return _LabeledWidget(title, widget, orientation=Qt.Orientation.Vertical)


def exec_formatter_dialog(ds: pd.Series, parent=None):
    dtype = ds.dtype
    if dtype.kind in "ui":
        dlg = QIntFormatterDialog(ds, parent=parent)
    elif dtype.kind == "f":
        dlg = QFloatFormatterDialog(ds, parent=parent)
    elif dtype.kind == "b":
        dlg = QBoolFormatterDialog(ds, parent=parent)
    elif dtype.kind == "c":
        dlg = QComplexFormatterDialog(ds, parent=parent)
    elif dtype.kind == "M":
        dlg = QDateTimeFormatterDialog(ds, parent=parent)
    elif dtype.kind == "m":
        dlg = QTimeDeltaFormatterDialog(ds, parent=parent)
    else:
        raise ValueError(f"Dtype {dtype.name} is not supported.")

    if dlg.exec():
        return dlg.toFormatter()
    return None



from ._item_model import AbstractDataFrameModel, DataFrameModel
from ._table_base import (
    QBaseTable,
    QMutableTable,
    QMutableSimpleTable,
)
from ._table_group import QTableGroup
from ._enhanced_table import _QTableViewEnhanced

# activate key combo
from . import _keycombo

del _keycombo

__all__ = [
    "AbstractDataFrameModel",
    "DataFrameModel",
    "QBaseTable",
    "QMutableTable",
    "QMutableSimpleTable",
    "_QTableViewEnhanced",
    "QTableGroup",
]



from __future__ import annotations

from typing import Any, Callable, Iterator, TYPE_CHECKING

from . import file, plot, column, selection, tab, table, view, analysis, window
from types import FunctionType, ModuleType
from qt_command_palette import get_palette

if TYPE_CHECKING:
    from tabulous._qt._mainwindow import _QtMainWidgetBase
    from tabulous.widgets import TableViewerBase

_SUB_MODULES: list[ModuleType] = [
    file,
    plot,
    column,
    selection,
    tab,
    table,
    view,
    analysis,
    window,
]

__all__ = [mod.__name__.split(".")[-1] for mod in _SUB_MODULES] + [
    "iter_commands",
    "register_command",
]


def iter_commands() -> Iterator[tuple[str, FunctionType]]:
    for mod in _SUB_MODULES:
        for obj in vars(mod).values():
            if isinstance(obj, FunctionType) and not obj.__name__.startswith("_"):
                yield mod.__name__.split(".")[-1], obj


def register_command(
    func: Callable = None,
    title: str = "User defined",
    desc: str = None,
) -> Callable[[Callable[[TableViewerBase], Any]], Callable[[TableViewerBase], Any]]:
    def wrapper(f):
        palette = get_palette("tabulous")
        if desc is None:
            fname = getattr(f, "__name__", None)
            if not isinstance(fname, str):
                raise TypeError(
                    f"Expected str for the function name, got {type(fname).__name__}"
                )
            _desc = fname.title().replace("_", " ")
        else:
            _desc = desc

        def fn(self: _QtMainWidgetBase):
            return f(self._table_viewer)

        fn.__doc__ = f.__doc__
        fn.__name__ = f.__name__
        palette.register(fn, title=title, desc=_desc)
        palette.update()
        return f

    return wrapper if func is None else wrapper(func)


DEFAULT_KEYBINDING_SETTING: list[tuple[FunctionType, str]] = [
    (window.toggle_toolbar, "Ctrl+K, Ctrl+T"),
    (window.toggle_console, "Ctrl+Shift+C"),
    (window.show_command_palette, ["F1", "Ctrl+Shift+P"]),
    (window.focus_table, "Ctrl+0"),
    (window.new_window, "Ctrl+Shift+N"),
    (window.toggle_fullscreen, "F11"),
    (window.close_window, "Ctrl+W"),
    (window.show_keymap, "Ctrl+K, Shift+?"),
    (window.toggle_focus, "Ctrl+Shift+F"),
    (window.show_preference, "Ctrl+,"),
    (tab.activate_left, "Alt+Left"),
    (tab.activate_right, "Alt+Right"),
    (tab.swap_tab_with_left, "Alt+Shift+Left"),
    (tab.swap_tab_with_right, "Alt+Shift+Right"),
    (tab.tile_with_adjacent_table, "Ctrl+K, ^"),
    (tab.untile_table, "Ctrl+K, \\"),
    (tab.delete_tab, "Ctrl+Delete"),
    (view.set_popup_mode, "Ctrl+K, P"),
    (view.set_dual_v_mode, "Ctrl+K, V"),
    (view.set_dual_h_mode, "Ctrl+K, H"),
    (view.reset_view_mode, "Ctrl+K, N"),
    (table.toggle_editability, "Ctrl+K, E"),
    (table.new_spreadsheet, "Ctrl+N"),
    (table.show_finder_widget, "Ctrl+F"),
    (table.show_undo_stack_view, "Ctrl+H"),
    (table.undo_table, "Ctrl+Z"),
    (table.redo_table, "Ctrl+Y"),
    (file.open_table, "Ctrl+O"),
    (file.open_spreadsheet, "Ctrl+K, Ctrl+O"),
    (file.save_table, "Ctrl+S"),
    (file.save_table_to_source, "Ctrl+Shift+S"),
    (selection.show_context_menu, "Menu"),
    (selection.raise_slot_error, "F6"),
    (selection.copy_data_tab_separated, "Ctrl+C"),
    (selection.copy_data_with_header_tab_separated, "Ctrl+C, Ctrl+H"),
    (selection.copy_as_new_table, "Ctrl+Shift+X"),
    (selection.select_all, "Ctrl+A"),
    (selection.cut_data, "Ctrl+X"),
    (selection.paste_data_tab_separated, "Ctrl+V"),
    (selection.delete_values, ["Delete", "Backspace"]),
]
